"""spec2nii module containing functions specific to interpreting the GE pfile format

Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2020 University of Oxford

This code is adapted from the VESPA project https://scion.duhs.duke.edu/vespa/project.
I therefore include their BSD statement here.

    Copyright (c) 2010, Duke University. All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:
        * Redistributions of source code must retain the above copyright notice,
          this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright
          notice, this list of conditions and the following disclaimer in the
          documentation and/or other materials provided with the distribution.
        * Neither the name of Duke University nor the United States Department
          of Veterans Affairs may be used to endorse or promote products derived
          from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ''AS
    IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
    PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE
    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
    SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
    OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
    DAMAGE.

    For portions of this code, copyright and license information differs from
    the above. In these cases, copyright and/or license information is inline.
"""

# Python modules
from datetime import datetime
from os.path import basename
from warnings import warn

# 3rd party modules
import numpy as np

from nifti_mrs.create_nmrs import gen_nifti_mrs_hdr_ext
from nifti_mrs.hdr_ext import Hdr_Ext

# Our modules
from spec2nii.GE.ge_read_pfile import Pfile
from spec2nii.nifti_orientation import NIFTIOrient
from spec2nii import __version__ as spec2nii_ver


class UnsupportedPulseSequenceError(Exception):
    """Raised when a user selected file does not contain the desired pulse sequence """
    pass


def read_pfile(filename, fname_out):
    """
    Given the name of a GE Pfile (*.7) file return a reference and metabolite
    data object.

    The mapper() method returns data as (x, y, z, navg, ncoil, nspectralpts)
    """

    pfile = Pfile(filename)

    if pfile.is_svs:
        data, fname_suffix = _process_svs_pfile(pfile)
    else:
        data, fname_suffix = _process_mrsi_pfile(pfile)

    fnames = []
    for fns in fname_suffix:
        if fname_out is None:
            fnames.append(filename.stem + fns)
        else:
            fnames.append(fname_out + fns)

    return data, fnames


def _process_svs_pfile(pfile):
    """Handle SVS data

    :param Pfile pfile: Pfile object
    :return: List of NIFTI MRS data objects
    :return: List of file name suffixes
    """
    psd = pfile.hdr.rhi_psdname.decode('utf-8').lower()
    proto = pfile.hdr.rhs_se_desc.decode('utf-8').lower()
    if psd == 'hbcd' and "press" in proto:
        print('\nPSD was: ', psd)
        print('Proto is: ', proto)
        psd = pfile.hdr.rhs_se_desc.decode('utf-8').lower()
        print('PSD updated to: ', psd)

    # MM: Some 'gaba' psd strings contain full path names, so truncate to the end of the path
    if psd.endswith('gaba'):
        psd = 'gaba'

    numecho = pfile.hdr.rhi_numecho

    if psd in ('mrs-press', 'probe-p', 'probe-s', 'probe-p_ach'):
        data, meta, dwelltime, fname_suffix = _process_probe_p(pfile)
    elif psd in ('oslaser', 'slaser_cni') and numecho == 1:  # MM: If non-edited data, use _process_oslaser
        data, meta, dwelltime, fname_suffix = _process_oslaser(pfile)
    elif psd == 'oslaser' and numecho > 1:  # MM: If edited data, use _process_gaba
        data, meta, dwelltime, fname_suffix = _process_gaba(pfile)
    elif psd == 'slaser':
        data, meta, dwelltime, fname_suffix = _process_slaser(pfile)
    elif psd in ('hbcd', 'jpress', 'jpress_ac', 'gaba', 'probe-p-mega_rml', 'repress7'):
        data, meta, dwelltime, fname_suffix = _process_gaba(pfile)
    elif psd in ('hbcd2'):                                     					# ATG
        data, meta, dwelltime, fname_suffix = _process_hbcd(pfile)              # ATG
    else:
        raise UnsupportedPulseSequenceError(f'Unrecognised sequence {psd}.')

    orientation = NIFTIOrient(_calculate_affine(pfile))

    out_nmrs = []
    for dd, mm in zip(data, meta):
        out_nmrs.append(gen_nifti_mrs_hdr_ext(dd, dwelltime, mm, orientation.Q44, no_conj=True))

    return out_nmrs, fname_suffix


def _process_probe_p(pfile):
    """Extract metabolite and reference data from a prob_p format pfile

    :param Pfile pfile: Pfile object
    :return: List numpy data arrays
    :return: List of file name suffixes
    """

    metab = pfile.map.raw_suppressed                    # typically (1,1,1,navg,ncoil,npts)
    metab = np.transpose(metab, [0, 1, 2, 5, 4, 3])     # swap to (1,1,1,npts,ncoil,navg)

    water = pfile.map.raw_unsuppressed                  # typically (1,1,1,navg,ncoil,npts)
    water = np.transpose(water, [0, 1, 2, 5, 4, 3])     # swap to (1,1,1,npts,ncoil,navg)

    dwelltime = 1 / pfile.hdr.rhr_spectral_width

    meta = _populate_metadata(pfile, water_suppressed=True)
    meta_ref = _populate_metadata(pfile, water_suppressed=False)

    meta.set_dim_info(0, 'DIM_COIL')
    meta.set_dim_info(1, 'DIM_DYN')

    meta_ref.set_dim_info(0, 'DIM_COIL')
    meta_ref.set_dim_info(1, 'DIM_DYN')

    return [metab, water], [meta, meta_ref], dwelltime, ['', '_ref']


def _process_oslaser(pfile):
    """Extract metabolite and reference data from a oslaser format pfile

    I think this is like the CMRR sLASER sequence with 2 ecc and 2 water scaling
    scans at the start and end of each acquisition.

    :param Pfile pfile: Pfile object
    :return: List numpy data arrays
    :return: List of file name suffixes
    """

    water = pfile.map.raw_unsuppressed                  # typically (1,1,1,navg,ncoil,npts)
    metab = pfile.map.raw_suppressed
    metab = np.transpose(metab, [0, 1, 2, 5, 4, 3])     # swap to (1,1,1,npts,ncoil,navg)
    water = np.transpose(water, [0, 1, 2, 5, 4, 3])     # swap to (1,1,1,npts,ncoil,navg)

    data = []
    meta = []
    fname_suffix = []
    data.append(water[:, :, :, :, :, [0, 1, 4, 5]])
    meta.append(_populate_metadata(pfile, water_suppressed=False, data_dimensions=data[0].ndim))
    fname_suffix.append('_quant')
    data.append(water[:, :, :, :, :, [2, 3, 6, 7]])
    meta.append(_populate_metadata(pfile, water_suppressed=False, data_dimensions=data[1].ndim))
    fname_suffix.append('_ecc')

    data.append(metab)
    meta.append(_populate_metadata(pfile, water_suppressed=True, data_dimensions=metab.ndim))
    fname_suffix.append('')

    dwelltime = 1 / pfile.hdr.rhr_spectral_width

    return data, meta, dwelltime, fname_suffix


def _process_slaser(pfile):
    """Extract metabolite and reference data from a slaser format pfile

    This seems to be like a standard probe-p. Maybe slaser is the canonical vendor implementation.

    :param Pfile pfile: Pfile object
    :return: List numpy data arrays
    :return: List of file name suffixes
    """

    metab = pfile.map.raw_suppressed                    # typically (1,1,1,navg,ncoil,npts)
    metab = np.transpose(metab, [0, 1, 2, 5, 4, 3])     # swap to (1,1,1,npts,ncoil,navg)

    water = pfile.map.raw_unsuppressed                  # typically (1,1,1,navg,ncoil,npts)
    water = np.transpose(water, [0, 1, 2, 5, 4, 3])     # swap to (1,1,1,npts,ncoil,navg)

    dwelltime = 1 / pfile.hdr.rhr_spectral_width

    meta = _populate_metadata(pfile, water_suppressed=True, data_dimensions=metab.ndim)
    meta_ref = _populate_metadata(pfile, water_suppressed=False, data_dimensions=water.ndim)

    return [metab, water], [meta, meta_ref], dwelltime, ['', '_ref']


def _add_editing_info(pfile, meta, data):
    """Add editing information to dimension tags and headers

    :param pfile: p-file object
    :type pfile: Pfile
    :param meta: Header extension object
    :type meta: Hdr_Ext
    :param data: Shaped complex data
    :type data: np.ndarray
    """
    edit_rf_waveform = pfile.hdr.rhi_user19
    # edit_rf_waveform == 19.0 is used by HERMES and HERCULES
    if data.shape[-1] == 2 and not edit_rf_waveform == 19.0:
        edit_rf_freq_off1 = pfile.hdr.rhi_user20
        edit_rf_freq_off2 = pfile.hdr.rhi_user21
        edit_rf_ppm_off1 = edit_rf_freq_off1 / float(pfile.hdr.rhr_rh_ps_mps_freq * 1E-7)
        edit_rf_ppm_off2 = edit_rf_freq_off2 / float(pfile.hdr.rhr_rh_ps_mps_freq * 1E-7)
        edit_rf_dur = pfile.hdr.rhi_user22
        # check for default value (-1) of pulse length
        if edit_rf_dur <= 0:
            edit_rf_dur = 16000
        dim_info = "MEGA-EDITED j-difference editing, two conditions"
        dim_header = {"EditCondition": ["ON", "OFF"]}
        edit_pulse_val = {
            "ON": {"PulseOffset": edit_rf_ppm_off1, "PulseDuration": edit_rf_dur / 1E6},
            "OFF": {"PulseOffset": edit_rf_ppm_off2, "PulseDuration": edit_rf_dur / 1E6}}

        meta.set_dim_info(2, 'DIM_EDIT', hdr=dim_header, info=dim_info)
        meta.set_standard_def("EditPulse", edit_pulse_val)
    else:
        meta.set_dim_info(2, 'DIM_EDIT')


def _process_gaba(pfile):
    """Extract metabolite and reference data from a gaba (MPRESS) format pfile

    :param Pfile pfile: Pfile object
    :return: List numpy data arrays
    :return: List of file name suffixes
    """

    # Note that custom mapper sorts dimensions already
    metab = pfile.map.raw_suppressed
    water = pfile.map.raw_unsuppressed

    dwelltime = 1 / pfile.hdr.rhr_spectral_width

    meta = _populate_metadata(pfile, water_suppressed=True)
    meta_ref = _populate_metadata(pfile, water_suppressed=False)

    meta.set_dim_info(0, 'DIM_COIL')
    meta.set_dim_info(1, 'DIM_DYN')
    # Only set an EDIT dim if there is an editing dimension
    if metab.ndim == 7:
        _add_editing_info(pfile, meta, metab)

    meta_ref.set_dim_info(0, 'DIM_COIL')
    meta_ref.set_dim_info(1, 'DIM_DYN')
    # Only set an EDIT dim if there is an editing dimension
    if water.ndim == 7:
        _add_editing_info(pfile, meta_ref, water)

    return [metab, water], [meta, meta_ref], dwelltime, ['', '_ref']


def _process_hbcd(pfile):
    """
    Input:
        Pfile Object

    Output:
        List of NumPy Data Arrays
        List of File Name Suffixes

    Details:
        Hyper/ISTHMUS Sequence

        The Integrated Short-TE and Hadamard-edited Multi-Sequence (ISTHMUS)
          incorporates a Short TE (35ms) PRESS, Long-TE (80ms) HERCULES, and
          a water reference for each.

        Data is organized within the file as follows:
          ( 1) Long  TE Reference : 80ms Unsupressed Water
          (32) Long  TE Edited    : 80ms Water Suppressed HERCULES
          ( 1) Short TE Reference : 35ms Unsupressed Water
          (32) Short TE Unedited  : 35ms Water Suppressed PRESS
          ( 1) Long  TE Reference : 80ms Unsupressed Water
          (32) Short TE Unedited  : 35ms Water Suppressed PRESS
          ( 1) Short TE Reference : 35ms Unsupressed Water
          (32) Short TE Unedited  : 35ms Water Suppressed PRESS
          ( 1) Long  TE Reference : 80ms Unsupressed Water
          (32) Short TE Unedited  : 35ms Water Suppressed PRESS
          ( 1) Short TE Reference : 35ms Unsupressed Water
          (32) Short TE Unedited  : 35ms Water Suppressed PRESS
          ( 1) Long  TE Reference : 80ms Unsupressed Water
          (32) Short TE Unedited  : 35ms Water Suppressed PRESS
          ( 1) Short TE Reference : 35ms Unsupressed Water
          (32) Short TE Unedited  : 35ms Water Suppressed PRESS

        Data is directly separated from the raw data (pfile.map.raw_data) where the data
          mapper (GABA mapper) is simply used to populate in the raw data.

        Author : Aaron Gudmundson, Johns Hopkins University, 2024
        Contact: agudmun2@jhmi.edu
    """

    # Additional Imports
    import copy

    # Editing Parameters
    edit_cases       = 4                                                                    # 4 Editing Conditions
    edit_pulse_1     = 4.58                                                                 # 4.58 ppm
    edit_pulse_2     = 1.90                                                                 # 1.90 ppm
    edit_pulse_4     = 4.18                                                                 # 4.18 ppm
    pulse_length     = 0.02                                                                 # Edit Pulse 20 ms

    dim_header       = {'EditCondition': ['A', 'B', 'C', 'D']}                              # 4 Subscans
    edit_pulse_val   = {'A': {'PulseOffset': [edit_pulse_1, edit_pulse_2], 'PulseDuration': pulse_length},
                        'B': {'PulseOffset': [edit_pulse_4, edit_pulse_2], 'PulseDuration': pulse_length},
                        'C': {'PulseOffset': edit_pulse_1, 'PulseDuration': pulse_length},
                        'D': {'PulseOffset': edit_pulse_4, 'PulseDuration': pulse_length}}

    # All Data (Skip 1st Transient - GE automatically has historically included a 'noise' transient)
    raw_data         = pfile.map.raw_data[:, :, :, :, 1:, :]                                # Raw Data from Mapper

    # Long TE HERCULES Metabolite Data
    lTE_metab        = copy.deepcopy(raw_data)                                              # Long TE Metab
    lTE_mask         = np.ones(lTE_metab.shape[4], dtype=bool)                              # Create a Mask
    lTE_mask[::33]   = False                                                                # Remove Water Refs
    lTE_mask[: 33]   = False                                                                # Remove PRESS
    lTE_metab        = lTE_metab[:, :, :, :, lTE_mask, :]                                   # Isolated HERCULES

    # Handle Incomplete
    if lTE_mask.shape[-1] % 4 != 0:                                                         # Incomplete Acquisition
        old_num_avgs = lTE_mask.shape[-1]                                                   # Old Total Averages
        new_num_avgs = (lTE_mask.shape[-1] // 4) * 4                                        # New Total Averages
        lTE_metab    = lTE_metab[:, :, :, :, :new_num_avgs, :]                              # Remove Incomplete

        notestring   = '80ms HERCULES'                                                      # Note Incomplete Data
        notestring   = f'{notestring} - Correcting - Incomplete Averages'                   # Note Incomplete Data
        notestring   = f'{notestring}  {old_num_avgs} --> {new_num_avgs}'                   # Note Incomplete Data
        print(f'{notestring} \t Corrected**')                                               # Note Incomplete Data

    new_shape     = list(lTE_metab.shape)                                                   # Remove Averages Dim
    new_shape[4]  = new_shape[4] // 4                                                       # Closest multiple of 4
    new_shape.append(edit_cases)                                                        	# Include Subscans

    lTE_metab_    = np.zeros(new_shape, dtype=np.complex128)                                # New lTE Metab Array
    lTE_metab_[:, :, :, :, :, :, 0] = lTE_metab[:, :, :, :, 0::4, :]                        # Subscan 1
    lTE_metab_[:, :, :, :, :, :, 1] = lTE_metab[:, :, :, :, 1::4, :]                        # Subscan 2
    lTE_metab_[:, :, :, :, :, :, 2] = lTE_metab[:, :, :, :, 2::4, :]                        # Subscan 3
    lTE_metab_[:, :, :, :, :, :, 3] = lTE_metab[:, :, :, :, 3::4, :]                        # Subscan 4
    lTE_metab      = lTE_metab_                                                             # With Subscan Dim

    lTE_metab_meta = _populate_metadata(pfile, water_suppressed=True)                       # Acquisition Information
    lTE_metab_meta.set_standard_def('EchoTime', 0.080)                                      # TE
    lTE_metab_meta.set_standard_def('WaterSuppressed', True)                                # Water Suppression
    lTE_metab_meta.set_standard_def('EditPulse', edit_pulse_val)                            # Header Edit Info

    lTE_metab_meta.set_dim_info(0, 'DIM_DYN')                                               # Dimension Info
    lTE_metab_meta.set_dim_info(1, 'DIM_COIL')                                              # Dimension Info
    lTE_metab_meta.set_dim_info(2, 'DIM_EDIT', hdr=dim_header)                              # Dimension Info

    # Short TE HERCULES Metabolite Data
    sTE_metab = copy.deepcopy(raw_data[:, :, :, :, 1:33, :])

    sTE_metab_meta = _populate_metadata(pfile, water_suppressed=True)                       # Acquisition Information
    sTE_metab_meta.set_standard_def('EchoTime', 0.035)                                      # TE
    sTE_metab_meta.set_standard_def('WaterSuppressed', True)                                # Water Suppression

    sTE_metab_meta.set_dim_info(0, 'DIM_DYN')                                               # Dimension Info
    sTE_metab_meta.set_dim_info(1, 'DIM_COIL')                                              # Dimension Info

    # Long TE Reference Water Data
    lTE_water = copy.deepcopy(raw_data[:, :, :, :, 0::66, :])

    lTE_water_meta = _populate_metadata(pfile, water_suppressed=False)                      # Acquisition Information
    lTE_water_meta.set_standard_def('EchoTime', 0.080)                                      # TE
    lTE_water_meta.set_standard_def('WaterSuppressed', False)                               # Water Suppression

    lTE_water_meta.set_dim_info(0, 'DIM_DYN')                                               # Dimension Info
    lTE_water_meta.set_dim_info(1, 'DIM_COIL')                                              # Dimension Info

    # Short TE Reference Water Data
    sTE_water = copy.deepcopy(raw_data[:, :, :, :, 33::66, :])

    sTE_water_meta = _populate_metadata(pfile, water_suppressed=False)                      # Acquisition Information
    sTE_water_meta.set_standard_def('EchoTime', 0.035)                                      # TE
    sTE_water_meta.set_standard_def('WaterSuppressed', False)                               # Water Suppression

    sTE_water_meta.set_dim_info(0, 'DIM_DYN')                                               # Dimension Info
    sTE_water_meta.set_dim_info(1, 'DIM_COIL')                                              # Dimension Info

    # Dwell Time
    dwelltime = 1 / pfile.hdr.rhr_spectral_width

    data      = [lTE_metab, sTE_metab, lTE_water, sTE_water]                                # ISTHMUS Data
    meta      = [lTE_metab_meta, sTE_metab_meta, lTE_water_meta, sTE_water_meta]            # ISTHMUS Header
    ref_names = ['_edited', '_short_te', '_ref_edited', '_ref_short_te']                    # ISTHMUS Naming

    print('Returning ISTHMUS Data:')
    for ii in range(len(data)):
        print('    {:02d} {:<14} '.format(ii, ref_names[ii]), data[ii].shape)
    print(' ')

    return data, meta, dwelltime, ref_names


def _process_mrsi_pfile(pfile):
    """Handle MRSI data

    :param Pfile pfile: Pfile object
    :return: List of NIFTI MRS data objects
    :return: List of file name suffixes
    """
    psd = pfile.hdr.rhi_psdname.decode('utf-8').lower()

    known_formats = ('probe-p', 'probe-sl', 'slaser_cni', 'presscsi')
    if psd not in known_formats:
        raise UnsupportedPulseSequenceError(
            f"Unrecognised sequence {psd}, psdname must be in: {','.join(known_formats)}.")

    warn('The interpretation of pfile CSI data is poorly tested; rotations or transpositions of the'
         ' CSI grid could be present. Spec2nii currently lacks a complete set of CSI test data.'
         ' Please get in touch to help solve this issue!')

    data = np.transpose(pfile.map.raw_data, [0, 1, 2, 5, 4, 3])
    if data.shape[5] == 1:
        data = data.squeeze(axis=5)

    # Perform fft
    def fft_and_shift(x, axis):
        return np.fft.fftshift(np.fft.fft(x, axis=axis), axes=axis)

    data = fft_and_shift(data, 0)
    data = fft_and_shift(data, 1)
    data = fft_and_shift(data, 2)

    dwelltime = 1 / pfile.hdr.rhr_spectral_width
    meta = _populate_metadata(pfile)
    meta.set_dim_info(0, 'DIM_COIL')

    orientation = NIFTIOrient(_calculate_affine_mrsi(pfile))

    return [gen_nifti_mrs_hdr_ext(data, dwelltime, meta, orientation.Q44, no_conj=True), ], ['', ]


def _calculate_affine_mrsi(pfile):
    """Calculate the 4x4 affine matrix for mrsi"""

    dcos = pfile.map.get_dcos.T
    dcos[dcos == 0.0] = 0.0             # remove -0.0 values

    voxel_size = pfile.map.get_voxel_spacing
    voxel_size_eye = np.diag(voxel_size)
    voi_position = pfile.map.get_origin_from_center(pfile.map.get_select_box_center,
                                                    pfile.map.get_num_voxels,
                                                    pfile.map.get_voxel_spacing,
                                                    pfile.map.get_dcos)

    affine = np.zeros((4, 4), dtype=float)
    affine[:3, :3] = dcos @ voxel_size_eye

    affine[:3, 3] = voi_position
    affine[3, 3] = 1.0

    return affine


def _calculate_affine(pfile):
    """Calculate the 4x4 affine matrix"""

    dcos = pfile.map.get_dcos.T
    dcos[dcos == 0.0] = 0.0             # remove -0.0 values

    voxel_size = pfile.map.get_select_box_size
    voxel_size_eye = np.diag(voxel_size)
    voi_position = pfile.map.get_select_box_center

    affine = np.zeros((4, 4), dtype=float)
    affine[:3, :3] = dcos @ voxel_size_eye

    affine[:3, 3] = voi_position
    affine[3, 3] = 1.0

    return affine


def _populate_metadata(pfile, water_suppressed=True, data_dimensions=None):
    """Populate a nifti-mrs header extension with metadata from the pfile

    If (up to 7) data_dimensions are specified then default dimension tags
    (coil, dyn, indirect) will be included. Otherwise manually specify
    outside this function.

    :param pfile: pfile object
    :type pfile: pfile map object
    :param water_suppressed: Set water suppression header field, defaults to True
    :type water_suppressed: bool, optional
    :param data_dimensions: If set to 5,6, or 7 will include default dim tags for those dimensions, defaults to None
    :type data_dimensions: int, optional
    :return: Header extension object
    :rtype: nifti_mrs.hdr_ext
    """
    hdr = pfile.hdr
    spec_frequency = float(pfile.hdr.rhr_rh_ps_mps_freq) / 1e7

    #  Use the mps freq and field strength to determine gamma and thus isotope
    try:
        gamma = (hdr.rhr_rh_ps_mps_freq * 1e-7) / (hdr.rhe_magstrength / 10000.0)
        if abs(gamma - 42.57) < 0.3:
            nucleus = "1H"
        elif abs(gamma - 10.7) < 0.3:
            nucleus = "13C"
        elif abs(gamma - 17.2) < 0.3:
            nucleus = "31P"
        else:
            print('Warning: Unrecognised nucleus, setting to 1H as default. '
                  'Use the --override_nucleus option to specify a different nuclide.')
            nucleus = "1H"
    except ZeroDivisionError:
        # Catch data (anonymised?) which has had rhe_magstrength set to 0
        # E.g. the BIG GABA data.
        print('Warning: Magnetic field value set to zero, setting nucleus to 1H as default. '
              'Use the --override_nucleus option to specify a different nuclide.')
        nucleus = "1H"

    meta = Hdr_Ext(
        spec_frequency,
        nucleus,
        dimensions=data_dimensions)

    # Standard defined metadata
    # # 5.1 MRS specific Tags
    # 'EchoTime'
    meta.set_standard_def('EchoTime', float(hdr.rhi_te) / 1E6)
    # 'RepetitionTime'
    meta.set_standard_def('RepetitionTime', float(hdr.rhi_tr) / 1E6)
    # 'InversionTime'
    meta.set_standard_def('InversionTime', float(hdr.rhi_ti) / 1E6)
    # 'MixingTime'
    # Not known
    # 'ExcitationFlipAngle'
    meta.set_standard_def('ExcitationFlipAngle', float(hdr.rhi_mr_flip))
    # 'TxOffset'
    # Not known
    # 'VOI'
    # Not known
    # 'WaterSuppressed'
    meta.set_standard_def('WaterSuppressed', water_suppressed)
    # 'WaterSuppressionType'
    # Not Known
    # 'SequenceTriggered'
    # Not Known

    # # 5.2 Scanner information
    # 'Manufacturer'
    meta.set_standard_def('Manufacturer', 'GE')
    # 'ManufacturersModelName'
    meta.set_standard_def('ManufacturersModelName', hdr.rhe_ex_sysid.decode('utf-8'))
    # 'DeviceSerialNumber'
    meta.set_standard_def('DeviceSerialNumber', hdr.rhe_uniq_sys_id.decode('utf-8'))
    # 'SoftwareVersions'
    meta.set_standard_def('SoftwareVersions', hdr.rhe_ex_verscre.decode('utf-8'))
    # 'InstitutionName'
    meta.set_standard_def('InstitutionName', hdr.rhe_hospname.decode('utf-8'))
    # 'InstitutionAddress'
    # Not known
    # 'TxCoil'
    # Not Known
    # 'RxCoil'
    meta.set_user_def(key='ReceiveCoilName', value=hdr.rhi_cname.decode('utf-8'), doc='Rx coil name.')

    # # 5.3 Sequence information
    # 'SequenceName'
    meta.set_standard_def('SequenceName', hdr.rhi_psdname.decode('utf-8'))
    # 'ProtocolName'
    meta.set_standard_def('ProtocolName', hdr.rhs_se_desc.decode('utf-8'))

    # # 5.4 Sequence information
    # 'PatientPosition'
    # Not known
    # 'PatientName'
    meta.set_standard_def('PatientName', hdr.rhe_patname.decode('utf-8'))
    # 'PatientID'
    # Not known
    # 'PatientWeight'
    # Not known
    # 'PatientDoB'
    meta.set_standard_def('PatientDoB', hdr.rhe_dateofbirth.decode('utf-8'))
    # 'PatientSex'
    if hdr.rhe_patsex == 1:
        sex_str = 'M'
    elif hdr.rhe_patsex == 2:
        sex_str = 'F'
    else:
        sex_str = 'O'
    meta.set_standard_def('PatientSex', sex_str)
    # # 5.5 Provenance and conversion metadata
    # 'ConversionMethod'
    meta.set_standard_def('ConversionMethod', f'spec2nii v{spec2nii_ver}')
    # 'ConversionTime'
    conversion_time = datetime.now().isoformat(sep='T', timespec='milliseconds')
    meta.set_standard_def('ConversionTime', conversion_time)
    # 'OriginalFile'
    meta.set_standard_def('OriginalFile', [basename(pfile.file_name)])
    # # 5.6 Spatial information
    # 'kSpace'
    meta.set_standard_def('kSpace', [False, False, False])

    return meta



"""spec2nii module containing functions specific to interpreting Philips DICOM
Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2020 University of Oxford
"""
from datetime import datetime

import numpy as np
import nibabel.nicom.dicomwrappers

from nifti_mrs.create_nmrs import gen_nifti_mrs_hdr_ext
from nifti_mrs.hdr_ext import Hdr_Ext

from spec2nii.dcm2niiOrientation.orientationFuncs import dcm_to_nifti_orientation
from spec2nii.nifti_orientation import NIFTIOrient
from spec2nii import __version__ as spec2nii_ver


class CSINotHandledError(Exception):
    pass


class IncorrectDataDimensionsError(Exception):
    pass


def _is_new_format(img):
    """Is the file old or new format DICOM"""
    media_storage_class = img.dcm_data.file_meta[('0002', '0002')].value
    if media_storage_class == '1.2.840.10008.5.1.4.1.1.4.2':
        return True
    elif media_storage_class == '1.3.46.670589.11.0.0.12.1':
        return False
    else:
        raise ValueError(f'Media Storage SOP Class UID {media_storage_class} not recognised as spectroscopy.')


def svs_or_CSI(img):
    """Identify from the headers whether data is CSI, SVS or FID."""
    if _is_new_format(img):
        if img.image_shape is None \
                and img.dcm_data.VolumeLocalizationTechnique == 'NONE':
            return 'FID'
        elif np.prod(img.image_shape) > 1.0:
            return 'CSI'
        else:
            return 'SVS'
    else:
        private_mrs_tags = img.dcm_data[('2005', '140f')][0]
        columns = private_mrs_tags[('0018', '9234')].value  # 'Spectroscopy Acquisition Phase Columns'
        rows = private_mrs_tags[('0018', '9095')].value  # 'Spectroscopy Acquisition Phase Rows'
        slices = private_mrs_tags[('0018', '9159')].value  # 'Spectroscopy Acquisition Phase Out-of-plane'

        if img.image_shape is None \
                and img.dcm_data.VolumeLocalizationTechnique == 'NONE':
            return 'FID'
        elif np.prod([columns, rows, slices]) > 1.0:
            return 'CSI'
        else:
            return 'SVS'


def multi_file_dicom(files_in, fname_out, tag, verbose):
    """Parse a list of Philips DICOM files"""

    # Convert each file (combine after)
    data_list = []
    ref_list = []
    orientation_list = []
    dwelltime_list = []
    meta_list = []
    meta_ref_list = []
    mainStr = ''
    for idx, fn in enumerate(files_in):
        if verbose:
            print(f'Converting dicom file {fn}')

        img = nibabel.nicom.dicomwrappers.wrapper_from_file(fn)

        mrs_type = svs_or_CSI(img)

        if mrs_type == 'SVS':
            specDataCmplx, spec_ref, orientation, dwelltime, meta_obj, meta_ref_obj = _process_philips_svs(img, verbose)

            newshape = (1, 1, 1) + specDataCmplx.shape
            spec_data = specDataCmplx.reshape(newshape)
            if spec_ref is not None:
                newshape = (1, 1, 1) + spec_ref.shape
                spec_ref = spec_ref.reshape(newshape)

            # Data appears to require conjugation to meet standard's conventions.
            spec_data = spec_data.conj()
            if spec_ref is not None:
                spec_ref = spec_ref.conj()

        elif mrs_type == 'FID':
            specDataCmplx, spec_ref, orientation, dwelltime, meta_obj, meta_ref_obj = _process_philips_fid(img, verbose)

            newshape = (1, 1, 1) + specDataCmplx.shape
            spec_data = specDataCmplx.reshape(newshape)
            spec_ref = spec_ref.reshape(newshape)

            # Data appears to require conjugation to meet standard's conventions.
            spec_data = spec_data.conj()
            spec_ref = spec_ref.conj()
        else:
            raise CSINotHandledError('CSI data is currently not handled for the Philips DICOM format.'
                                     'Please contact the developers if you have examples of this type of data.')

        data_list.append(spec_data)
        ref_list.append(spec_ref)
        orientation_list.append(orientation)
        dwelltime_list.append(dwelltime)
        meta_list.append(meta_obj)
        meta_ref_list.append(meta_ref_obj)

        if idx == 0:
            if fname_out:
                mainStr = fname_out
            else:
                mainStr = img.dcm_data.SeriesDescription

    # If data shape, orientation and dwelltime match combine
    # into one NIFTI MRS object.
    # Otherwise return a list of files/names
    def all_equal(lst):
        return lst[:-1] == lst[1:]

    combine = all_equal([d.shape for d in data_list])\
        and all_equal([o.Q44.tolist() for o in orientation_list])\
        and all_equal(dwelltime_list)

    nifti_mrs_out, fnames_out = [], []
    if combine:
        # Combine files into single MRS NIfTI
        # Single file name
        fnames_out.append(mainStr)
        if ref_list[0] is not None:
            fnames_out.append(mainStr + '_ref')

        dt_used = dwelltime_list[0]
        or_used = orientation_list[0]

        # Add original files to nifti meta information.
        meta_used = meta_list[0]
        meta_used.set_standard_def('OriginalFile', [str(ff.name) for ff in files_in])
        if meta_ref_list[0] is not None:
            meta_ref_used = meta_ref_list[0]
            meta_ref_used.set_standard_def('OriginalFile', [str(ff.name) for ff in files_in])

        # Combine data into 5th dimension if needed
        if len(data_list) > 1:
            combined_data = np.stack(data_list, axis=-1)
            combined_ref = np.stack(ref_list, axis=-1)
        else:
            combined_data = data_list[0]
            combined_ref = ref_list[0]

        # Add dimension information (if not None for default)
        if tag:
            meta_used.set_dim_info(0, tag)

        # Create NIFTI MRS object.
        nifti_mrs_out.append(gen_nifti_mrs_hdr_ext(combined_data, dt_used, meta_used, or_used.Q44, no_conj=True))
        if ref_list[0] is not None:
            nifti_mrs_out.append(gen_nifti_mrs_hdr_ext(combined_ref, dt_used, meta_ref_used, or_used.Q44, no_conj=True))
    else:
        for idx, (dd, rr, oo, dt, mm, mr, ff) in enumerate(zip(data_list,
                                                               ref_list,
                                                               orientation_list,
                                                               dwelltime_list,
                                                               meta_list,
                                                               meta_ref_list,
                                                               files_in)):
            # Add original files to nifti meta information.
            mm.set_standard_def('OriginalFile', [str(ff.name), ])
            fnames_out.append(f'{mainStr}_{idx:03}')
            nifti_mrs_out.append(gen_nifti_mrs_hdr_ext(dd, dt, mm, oo.Q44, no_conj=True))

            if rr is not None:
                mr.set_standard_def('OriginalFile', [str(ff.name), ])
                fnames_out.append(f'{mainStr}_ref_{idx:03}')
                nifti_mrs_out.append(gen_nifti_mrs_hdr_ext(rr, dt, mr, oo.Q44, no_conj=True))

    return nifti_mrs_out, fnames_out


def _process_philips_svs(img, verbose):
    """Process svs data using, using the relevant pathway"""
    if _is_new_format(img):
        return _process_philips_svs_new(img, verbose)
    else:
        return _process_philips_svs_old(img, verbose)


def _process_philips_svs_old(img, verbose):
    """Process Philips DICOM SVS data in the 'old' DICOM format"""
    specData = np.frombuffer(img.dcm_data[('2005', '1270')].value, dtype=np.single)
    specDataCmplx = specData[0::2] + 1j * specData[1::2]

    # 1) Extract dicom parameters
    currNiftiOrientation = _enhanced_dcm_svs_to_orientation(img, verbose)

    dwelltime = 1.0 / img.dcm_data[('2005', '1357')].value
    meta = _extractDicomMetadata_old(img)

    return specDataCmplx, None, currNiftiOrientation, dwelltime, meta, None


def _process_philips_svs_new(img, verbose):
    """Process Philips DICOM SVS data in the 'new' DICOM format"""
    specData = np.frombuffer(img.dcm_data[('5600', '0020')].value, dtype=np.single)
    specDataCmplx = specData[0::2] + 1j * specData[1::2]

    # (0028,0008) IS [32]                                               # 2,1 Number of Frames
    nframes = int(img.dcm_data[0x0028, 0x0008].value)
    spec_points = img.dcm_data.SpectroscopyAcquisitionDataColumns

    npoints_expected = spec_points * nframes
    if npoints_expected != specDataCmplx.size:
        raise IncorrectDataDimensionsError(
            f'Number of points {specDataCmplx.size}, does not equal frames ({nframes}) '
            f'* spectral points ({spec_points}).')

    # Only two conditions are known for the Spectroscopy scans 'normal/SV_PRESS' and 'SV_MEGA'
    # So special case the MEGA and everything else goes through the PRESS pipeline.
    try:
        is_edited = img.dcm_data[0x2005, 0x1597].value == 'Y'
    except KeyError:
        is_edited = False

    if img.dcm_data[0x0008, 0x9209].value == 'SPECTROSCOPY'\
            and is_edited:
        # Data is MEGA
        on_idx = []
        off_idx = []
        ref_idx = []
        for idx, instance in enumerate(img.dcm_data[0x2005, 0x140f]):
            if instance[0x2005, 0x1304].value:
                ref_idx.append(idx)
            else:
                if instance[0x2005, 0x1598].value == '1':
                    on_idx.append(idx)
                elif instance[0x2005, 0x1598].value == '0':
                    off_idx.append(idx)

        specDataCmplx = specDataCmplx.reshape((nframes, spec_points))

        spec_data_main = np.stack(
            (specDataCmplx[on_idx, :].T,
             specDataCmplx[off_idx, :].T), axis=-1)
        spec_data_ref = specDataCmplx[ref_idx, :].T

        meta = _extractDicomMetadata_new(img)
        meta_r = _extractDicomMetadata_new(img, water_suppressed=False)

        meta.set_dim_info("5th", "DIM_DYN")
        meta.set_dim_info("6th", "DIM_EDIT", hdr={"EditCondition": ["ON", "OFF"]})

        meta_r.set_dim_info("5th", "DIM_DYN")

    else:
        # In the one piece of data I have been provided the data is twice as long as indicated (1 avg)
        # and the second half is a water reference.

        # This actually tells you if it is selected in the UI, but seems to be a surrogate
        # for acquired in the PRESS scan
        reference_inlcuded = img.dcm_data[0x0018, 0x9199].value == 'YES'

        if reference_inlcuded:
            spec_data_main = specDataCmplx[:spec_points]
            spec_data_ref = specDataCmplx[spec_points:]
            spec_data_main = spec_data_main.reshape(nframes - 1, spec_points).T
            spec_data_main = spec_data_main.squeeze()
        else:
            spec_data_main = specDataCmplx.reshape(nframes, spec_points).T
            spec_data_main = spec_data_main.squeeze()
            spec_data_ref = None

        meta = _extractDicomMetadata_new(img)
        if reference_inlcuded:
            meta_r = _extractDicomMetadata_new(img, water_suppressed=False)
        else:
            meta_r = None

    currNiftiOrientation = _enhanced_dcm_svs_to_orientation(img, verbose)
    dwelltime = 1.0 / img.dcm_data.SpectralWidth

    return spec_data_main, spec_data_ref, currNiftiOrientation, dwelltime, meta, meta_r


def _process_philips_fid(img, verbose):
    """Process Philips DICOM FID data"""

    specData = np.frombuffer(img.dcm_data[('5600', '0020')].value, dtype=np.single)
    specDataCmplx = specData[0::2] + 1j * specData[1::2]

    # In the one piece of data I have been provided the data is twice as long as indicated (1 avg)
    # and the second half is a water reference.
    spec_points = img.dcm_data.SpectroscopyAcquisitionDataColumns
    spec_data_main = specDataCmplx[:spec_points]
    spec_data_ref = specDataCmplx[spec_points:]

    # 1) Extract dicom parameters
    defaultaffine = np.diag(np.array([10000, 10000, 10000, 1]))
    currNiftiOrientation = NIFTIOrient(defaultaffine)

    dwelltime = 1.0 / img.dcm_data.SpectralWidth
    meta = _extractDicomMetadata_new(img)
    meta_r = _extractDicomMetadata_new(img, water_suppressed=False)

    return spec_data_main, spec_data_ref, currNiftiOrientation, dwelltime, meta, meta_r


def _enhanced_dcm_svs_to_orientation(img, verbose=False):
    '''Convert the Volume Localization Sequence (0018,9126) enhanced DICOM tag
    to a 4x4 affine format.

    Assumes the center of all slabs are aligned currently.'''

    # Extract dicom parameters
    if _is_new_format(img):
        # Enhanced DICOM

        '''
        # Code before July 2023 when testing using JH data revealed problems.
        imageOrientationPatient = img.dcm_data.PerFrameFunctionalGroupsSequence[0]\
            .PlaneOrientationSequence[0].ImageOrientationPatient
        imageOrientationPatient = np.asarray(imageOrientationPatient).reshape(2, 3)
        imagePositionPatient = img.dcm_data.PerFrameFunctionalGroupsSequence[0]\
            .PlanePositionSequence[0].ImagePositionPatient
        imagePositionPatient = np.asarray(imagePositionPatient)
        xyzMM = [float(img.dcm_data.PerFrameFunctionalGroupsSequence[0].PixelMeasuresSequence[0].PixelSpacing[0]),
                 float(img.dcm_data.PerFrameFunctionalGroupsSequence[0].PixelMeasuresSequence[0].PixelSpacing[1]),
                 float(img.dcm_data.PerFrameFunctionalGroupsSequence[0].PixelMeasuresSequence[0].SliceThickness)]
        xyzMM = np.asarray(xyzMM)
        '''

        locSeq = img.dcm_data.VolumeLocalizationSequence
        imageOrientationPatient = np.concatenate((locSeq[0].SlabOrientation, locSeq[1].SlabOrientation))
        imageOrientationPatient = np.asarray(imageOrientationPatient).reshape(2, 3)
        imageOrientationPatient *= -1
        imagePositionPatient = np.asarray(locSeq[0].MidSlabPosition)
        xyzMM = np.asarray(
            [locSeq[1].SlabThickness,
             locSeq[0].SlabThickness,
             locSeq[2].SlabThickness])

        return dcm_to_nifti_orientation(
            imageOrientationPatient,
            imagePositionPatient,
            xyzMM,
            (1, 1, 1),
            half_shift=False,
            verbose=verbose)
    else:
        # Classic DICOM
        import warnings
        warnings.warn('The orientation and position code for old format Philips DICOM is untested.\
                       Please provide example data to improve the handling of this format.')
        # As implemented in vespa (vespa/analysis/fileio/dicom_philips.py#L181)
        try:
            section  = img.dcm_data[0x2005, 0x1085][0]
            angle_lr = float(section[0x02005, 0x1056].value)
            angle_ap = float(section[0x02005, 0x1054].value)
            angle_hf = float(section[0x02005, 0x1055].value)
            dim_lr   = float(section[0x02005, 0x1059].value)
            dim_ap   = float(section[0x02005, 0x1057].value)
            dim_hf   = float(section[0x02005, 0x1058].value)
            shift_lr = float(section[0x02005, 0x105c].value)
            shift_ap = float(section[0x02005, 0x105a].value)
            shift_hf = float(section[0x02005, 0x105b].value)

            # This matches the spar/sdat pathway in spec2nii.Philips.philips._philips_orientation
            from spec2nii.nifti_orientation import NIFTIOrient, calc_affine
            angles = [-angle_lr, -angle_ap, angle_hf]
            dimensions = [dim_lr, dim_ap, dim_hf]
            shift = [-shift_lr, -shift_ap, shift_hf]

            return NIFTIOrient(calc_affine(angles, dimensions, shift))

        except KeyError:
            # Default orientation
            print(
                'No orientation information found in header. Tag (2005, 1085) missing. '
                'Default orientation will be used.')
            imageOrientationPatient = [[1, 0, 0], [0, 1, 0]]
            imagePositionPatient = [0, 0, 0]
            xyzMM = [float(x) for x in img.dcm_data[0x0028, 0x0030].value] \
                + [float(img.dcm_data[0x0018, 0x0050].value), ]

            return dcm_to_nifti_orientation(
                imageOrientationPatient,
                imagePositionPatient,
                xyzMM,
                (1, 1, 1),
                half_shift=False,
                verbose=verbose)


def _extractDicomMetadata_old(dcmdata, water_suppressed=True):
    """ Extract information from the nibabel DICOM object to insert into the json header ext.

    Args:
        dcmdata: nibabel.nicom image object
    Returns:
        obj (hdr_ext): NIfTI MRS hdr ext object.
    """

    # Extract required metadata and create hdr_ext object
    try:
        frequency = dcmdata.dcm_data.TransmitterFrequency
    except AttributeError:
        frequency = float(dcmdata.dcm_data[('2001', '1083')].value)

    try:
        nucleus = dcmdata.dcm_data.ResonantNucleus
    except AttributeError:
        nucleus = dcmdata.dcm_data[('2001', '1087')].value

    obj = Hdr_Ext(
        frequency,
        nucleus)

    # private_mrs_tags = dcmdata[('2005', '140f')][0]

    def set_if_present(tag, val):
        if val:
            obj.set_standard_def(tag, val)

    # Standard metadata
    # # 5.1 MRS specific Tags
    # 'EchoTime'
    try:
        echo_time = float(dcmdata.dcm_data.PerFrameFunctionalGroupsSequence[0]
                          .MREchoSequence[0].EffectiveEchoTime) * 1E-3
    except AttributeError:
        echo_time = float(dcmdata.dcm_data[('2005', '1310')].value)
    obj.set_standard_def('EchoTime', echo_time)

    # 'RepetitionTime'
    try:
        rep_time = float(dcmdata.dcm_data.PerFrameFunctionalGroupsSequence[0]
                         .MRTimingAndRelatedParametersSequence[0].RepetitionTime) / 1E3
    except AttributeError:
        rep_time = dcmdata.dcm_data.RepetitionTime
    obj.set_standard_def('RepetitionTime', rep_time)

    # 'InversionTime'
    # Not known
    # 'MixingTime'
    # Not known
    # 'ExcitationFlipAngle'
    try:
        fa = float(dcmdata.dcm_data.PerFrameFunctionalGroupsSequence[0]
                   .MRTimingAndRelatedParametersSequence[0].RepetitionTime)
    except AttributeError:
        fa = dcmdata.dcm_data.FlipAngle
    obj.set_standard_def('ExcitationFlipAngle', fa)

    # 'TxOffset'
    # Not known
    # 'VOI'
    # Not known
    # 'WaterSuppressed'
    obj.set_standard_def('WaterSuppressed', water_suppressed)
    # 'WaterSuppressionType'
    # Not known
    # 'SequenceTriggered'
    # Not known
    # # 5.2 Scanner information
    # 'Manufacturer'
    obj.set_standard_def('Manufacturer', dcmdata.dcm_data.Manufacturer)
    # 'ManufacturersModelName'
    obj.set_standard_def('ManufacturersModelName', dcmdata.dcm_data.ManufacturerModelName)
    # 'DeviceSerialNumber'
    if 'DeviceSerialNumber' in dcmdata.dcm_data:
        obj.set_standard_def('DeviceSerialNumber', str(dcmdata.dcm_data.DeviceSerialNumber))
    # 'SoftwareVersions'
    obj.set_standard_def('SoftwareVersions', str(dcmdata.dcm_data.SoftwareVersions))
    # 'InstitutionName'
    obj.set_standard_def('InstitutionName', dcmdata.dcm_data.InstitutionName)
    # 'InstitutionAddress'
    if 'InstitutionAddress' in dcmdata.dcm_data:
        obj.set_standard_def('InstitutionAddress', dcmdata.dcm_data.InstitutionAddress)
    # 'TxCoil'
    # Not known
    # 'RxCoil'
    # ToDo
    # img.dcm_data.SharedFunctionalGroupsSequence[0].MRReceiveCoilSequence[0].ReceiveCoilName

    # # 5.3 Sequence information
    # 'SequenceName'
    if 'SequenceName' in dcmdata.dcm_data:
        obj.set_standard_def('SequenceName', dcmdata.dcm_data.PulseSequenceName)
    # 'ProtocolName'
    obj.set_standard_def('ProtocolName', dcmdata.dcm_data.ProtocolName)
    # # 5.4 Sequence information
    # 'PatientPosition'
    obj.set_standard_def('PatientPosition', dcmdata.dcm_data.PatientPosition)
    # 'PatientName'
    if dcmdata.dcm_data.PatientName:
        obj.set_standard_def('PatientName', dcmdata.dcm_data.PatientName.family_name)
    # 'PatientID'
    # Not known
    # 'PatientWeight'
    if 'PatientWeight' in dcmdata.dcm_data and dcmdata.dcm_data.PatientWeight:
        obj.set_standard_def('PatientWeight', float(dcmdata.dcm_data.PatientWeight))
    # 'PatientDoB'
    set_if_present('PatientDoB', dcmdata.dcm_data.PatientBirthDate)
    # 'PatientSex'
    set_if_present('PatientSex', dcmdata.dcm_data.PatientSex)

    # # 5.5 Provenance and conversion metadata
    # 'ConversionMethod'
    obj.set_standard_def('ConversionMethod', f'spec2nii v{spec2nii_ver}')
    # 'ConversionTime'
    conversion_time = datetime.now().isoformat(sep='T', timespec='milliseconds')
    obj.set_standard_def('ConversionTime', conversion_time)
    # 'OriginalFile'
    # Added later
    # # 5.6 Spatial information
    # 'kSpace'
    obj.set_standard_def('kSpace', [False, False, False])

    return obj


def _extractDicomMetadata_new(dcmdata, water_suppressed=True):
    """ Extract information from the nibabel DICOM object to insert into the json header ext.

    Args:
        dcmdata: nibabel.nicom image object
    Returns:
        obj (hdr_ext): NIfTI MRS hdr ext object.
    """

    # Extract required metadata and create hdr_ext object
    obj = Hdr_Ext(
        dcmdata.dcm_data.TransmitterFrequency,
        dcmdata.dcm_data.ResonantNucleus)

    def set_if_present(tag, val):
        if val:
            obj.set_standard_def(tag, val)

    # Standard metadata
    # # 5.1 MRS specific Tags
    # 'EchoTime'
    echo_time = float(dcmdata.dcm_data.PerFrameFunctionalGroupsSequence[0]
                      .MREchoSequence[0].EffectiveEchoTime) * 1E-3
    obj.set_standard_def('EchoTime', echo_time)
    # 'RepetitionTime'
    try:
        rep_tim = float(dcmdata.dcm_data.PerFrameFunctionalGroupsSequence[0]
                        .MRTimingAndRelatedParametersSequence[0].RepetitionTime) / 1E3
    except AttributeError:
        rep_tim = float(dcmdata.dcm_data.SharedFunctionalGroupsSequence[0]
                        .MRTimingAndRelatedParametersSequence[0].RepetitionTime) / 1E3

    obj.set_standard_def('RepetitionTime', rep_tim)
    # 'InversionTime'
    # Not known
    # 'MixingTime'
    # Not known
    # 'ExcitationFlipAngle'
    try:
        fa = float(dcmdata.dcm_data.PerFrameFunctionalGroupsSequence[0]
                   .MRTimingAndRelatedParametersSequence[0].FlipAngle)
    except AttributeError:
        fa = float(dcmdata.dcm_data.SharedFunctionalGroupsSequence[0]
                   .MRTimingAndRelatedParametersSequence[0].FlipAngle)
    obj.set_standard_def('ExcitationFlipAngle', fa)
    # 'TxOffset'
    # Not known
    # 'VOI'
    # Not known
    # 'WaterSuppressed'
    obj.set_standard_def('WaterSuppressed', water_suppressed)
    # 'WaterSuppressionType'
    # Not known
    # 'SequenceTriggered'
    # Not known
    # # 5.2 Scanner information
    # 'Manufacturer'
    obj.set_standard_def('Manufacturer', dcmdata.dcm_data.Manufacturer)
    # 'ManufacturersModelName'
    obj.set_standard_def('ManufacturersModelName', dcmdata.dcm_data.ManufacturerModelName)
    # 'DeviceSerialNumber'
    if 'DeviceSerialNumber' in dcmdata.dcm_data:
        obj.set_standard_def('DeviceSerialNumber', str(dcmdata.dcm_data.DeviceSerialNumber))
    # 'SoftwareVersions'
    obj.set_standard_def('SoftwareVersions', str(dcmdata.dcm_data.SoftwareVersions))
    # 'InstitutionName'
    if 'InstitutionName' in dcmdata.dcm_data:
        obj.set_standard_def('InstitutionName', dcmdata.dcm_data.InstitutionName)
    # 'InstitutionAddress'
    if 'InstitutionAddress' in dcmdata.dcm_data:
        obj.set_standard_def('InstitutionAddress', dcmdata.dcm_data.InstitutionAddress)
    # 'TxCoil'
    # Not known
    # 'RxCoil'
    # ToDo
    # img.dcm_data.SharedFunctionalGroupsSequence[0].MRReceiveCoilSequence[0].ReceiveCoilName

    # # 5.3 Sequence information
    # 'SequenceName'
    obj.set_standard_def('SequenceName', dcmdata.dcm_data.PulseSequenceName)
    # 'ProtocolName'
    if 'ProtocolName' in dcmdata.dcm_data:
        obj.set_standard_def('ProtocolName', dcmdata.dcm_data.ProtocolName)
    # # 5.4 Sequence information
    # 'PatientPosition'
    obj.set_standard_def('PatientPosition', dcmdata.dcm_data.PatientPosition)
    # 'PatientName'
    if dcmdata.dcm_data.PatientName:
        obj.set_standard_def('PatientName', dcmdata.dcm_data.PatientName.family_name)
    # 'PatientID'
    # Not known
    # 'PatientWeight'
    if 'PatientWeight' in dcmdata.dcm_data and dcmdata.dcm_data.PatientWeight:
        obj.set_standard_def('PatientWeight', float(dcmdata.dcm_data.PatientWeight))
    # 'PatientDoB'
    set_if_present('PatientDoB', dcmdata.dcm_data.PatientBirthDate)
    # 'PatientSex'
    set_if_present('PatientSex', dcmdata.dcm_data.PatientSex)

    # # 5.5 Provenance and conversion metadata
    # 'ConversionMethod'
    obj.set_standard_def('ConversionMethod', f'spec2nii v{spec2nii_ver}')
    # 'ConversionTime'
    conversion_time = datetime.now().isoformat(sep='T', timespec='milliseconds')
    obj.set_standard_def('ConversionTime', conversion_time)
    # 'OriginalFile'
    # Added later
    # # 5.6 Spatial information
    # 'kSpace'
    obj.set_standard_def('kSpace', [False, False, False])

    return obj



'''Tests for jmrui format conversion.
No spatial information is stored in jMRUI files.

Copyright William Clarke, University of Oxford 2021
Subject to the BSD 3-Clause License.
'''
import subprocess
from pathlib import Path

import pytest
import numpy as np
import os.path as op
import nibabel as nib

file_path = Path(__file__).parent
testdata = {'fida_single': file_path / 'spec2nii_test_data/jmrui/metab_jmrui.txt',
            'txt_multi': file_path / 'spec2nii_test_data/jmrui/basis_woTMS.txt',
            'mrui_single': file_path / 'spec2nii_test_data/jmrui/press2.mrui',
            'mrui_multi': file_path / 'spec2nii_test_data/jmrui/ALLSIG.mrui',
            'txt_old': file_path / 'spec2nii_test_data/jmrui/P31rest_proc_jmrui.txt'}


@pytest.fixture
def affine_file(tmp_path):

    affinepath = op.join(tmp_path, 'affine.txt')

    affine = np.random.random((4, 4))
    affine[3, 0] = 0.0
    affine[3, 1] = 0.0
    affine[3, 2] = 0.0
    affine[3, 3] = 1.0

    np.savetxt(affinepath, affine)

    return affinepath


def test_jmrui_single_txt(affine_file, tmp_path):
    """Test the 'jmrui' text format conversion on a single fid written from
    FID-A.
    """
    # Run spec2nii on text
    subprocess.call(['spec2nii', 'jmrui',
                     '-f', 'fida_single',
                     '-o', str(tmp_path),
                     '-a', affine_file,
                     '-j', testdata['fida_single']])

    # Load the new nifti file
    converted = nib.load(tmp_path / 'fida_single.nii.gz')

    assert converted.shape == (1, 1, 1, 4096)
    assert np.iscomplexobj(converted.dataobj)
    assert np.allclose(np.loadtxt(affine_file), converted.affine)


def test_jmrui_multi_txt(affine_file, tmp_path):
    """Test the 'jmrui' text format conversion on a multiple fids.
    """
    # Run spec2nii on text
    subprocess.call(['spec2nii', 'jmrui',
                     '-f', 'txt_multi',
                     '-o', str(tmp_path),
                     '-a', affine_file,
                     '-j', testdata['txt_multi']])

    # Load the new nifti file
    converted = nib.load(tmp_path / 'txt_multi.nii.gz')

    assert converted.shape == (1, 1, 1, 2048, 22)
    assert np.iscomplexobj(converted.dataobj)
    assert np.allclose(np.loadtxt(affine_file), converted.affine)


def test_jmrui_single_mrui(affine_file, tmp_path):
    """Test the 'jmrui' mrui format conversion on a single fid.
    """
    # Run spec2nii on text
    subprocess.call(['spec2nii', 'jmrui',
                     '-f', 'mrui_single',
                     '-o', str(tmp_path),
                     '-a', affine_file,
                     '-j', testdata['mrui_single']])

    # Load the new nifti file
    converted = nib.load(tmp_path / 'mrui_single.nii.gz')

    assert converted.shape == (1, 1, 1, 2048)
    assert np.iscomplexobj(converted.dataobj)
    assert np.allclose(np.loadtxt(affine_file), converted.affine)


def test_jmrui_multi_mrui(affine_file, tmp_path):
    """Test the 'jmrui' mrui format conversion on a multiple fids.
    """
    # Run spec2nii on text
    subprocess.call(['spec2nii', 'jmrui',
                     '-f', 'mrui_multi',
                     '-o', str(tmp_path),
                     '-a', affine_file,
                     '-j', testdata['mrui_multi']])

    # Load the new nifti file
    converted = nib.load(tmp_path / 'mrui_multi.nii.gz')

    assert converted.shape == (1, 1, 1, 2048, 28)
    assert np.iscomplexobj(converted.dataobj)
    assert np.allclose(np.loadtxt(affine_file), converted.affine)


def test_jmrui_old(affine_file, tmp_path):
    """Test the 'jmrui' txt old format with different # signal syntax.
    Namely, 'Signal number: ' rather than 'Signal')
    """
    # Run spec2nii on text
    subprocess.call(['spec2nii', 'jmrui',
                     '-f', 'txt_old',
                     '-o', str(tmp_path),
                     '-a', affine_file,
                     '-j', testdata['txt_old']])

    # Load the new nifti file
    converted = nib.load(tmp_path / 'txt_old.nii.gz')

    assert converted.shape == (1, 1, 1, 2048)
    assert np.iscomplexobj(converted.dataobj)
    assert np.allclose(np.loadtxt(affine_file), converted.affine)



"""spec2nii module wrapping an importer for the Varian file format

"Author": Jack J. Miller, jack.miller@physics.org
Copyright (C) 2021 University of Oxford and University of Aarhus
"""

import re
import warnings
from datetime import datetime
from os.path import basename, splitext

import numpy as np

from nifti_mrs.create_nmrs import gen_nifti_mrs_hdr_ext
from nifti_mrs.hdr_ext import Hdr_Ext

from spec2nii import __version__ as spec2nii_ver
from spec2nii.nifti_orientation import NIFTIOrient, calc_affine
import spec2nii.varian as v


def read_varian(args):
    """read_varian -- load a varian fid.
    Note that this format is very flexible and some rather large assumptions are made.
    At the moment, this assumes little.
    :param file: path to the varian .fid directory, containing fid and procpar files.
    returns img_out, fname_out
    """
    dic, data = v.read(args.file)

    # extract number of coils
    number_of_coils = 0
    for i in dic['procpar']['rcvrs']['values'][0]:
        if re.search(i, 'y'):
            number_of_coils += 1

    # number of time points -- probably
    number_of_time_points = float(dic['procpar']['arraydim']['values'][0]) / number_of_coils
    if (not number_of_time_points.is_integer()):
        raise ValueError('Coil reshaping failed')

    number_of_time_points = int(number_of_time_points)

    # spectral number of points
    number_of_spectral_points = int(int(dic['np']) / 2)

    # number of slices
    sequence_name = dic['procpar']['seqfil']['values'][0]
    if (sequence_name.count('SSel') or sequence_name.count('sliceshim')):  # 1D localised slice
        pass  # to be created

    # reshape
    newshape = (1, 1, 1, number_of_spectral_points, number_of_coils, number_of_time_points)
    data = data.transpose()
    data = np.resize(data, newshape)

    # extract additional spectral metadata
    dwelltime = 1.0 / float(dic['procpar']['sw']['values'][0])
    imagingfreq = float(dic['procpar']['sfrq']['values'][0])
    nucleus = dic['procpar']['tn']['values'][0]
    # reshape to be in the form '13C' rather than 'C13'
    nucleus = nucleus[1:] + nucleus[0]

    try:
        repetition_time = float(dic['procpar']['tr']['values'][0])
    except KeyError:
        pass

    try:
        echotime = float(dic['procpar']['te']['values'][0])  # In ms if 'tis there
    except KeyError:
        pass
    try:
        echotime = float(dic['procpar']['pw']['values'][0]) + float(dic['procpar']['alfa']['values'][0])
    except KeyError:
        pass
    try:
        echotime = float(dic['procpar']['p1']['values'][0]) + float(dic['procpar']['alfa']['values'][0])
    except KeyError:
        pass

    # Parse 3D localisation
    if (sequence_name.count('press') or sequence_name.count('steam')):
        affine = _varian_orientation_3d(dic)
    else:
        affine = np.diag(np.array([10000, 10000, 10000, 1]))  # 10 m slab for now....

    # TODO: Jack should implement the affine matrix correctly for all sequences

    orientation = NIFTIOrient(affine)

    # create object
    meta = Hdr_Ext(imagingfreq, nucleus)
    meta.set_standard_def('ConversionMethod', f'spec2nii v{spec2nii_ver}')
    meta.set_standard_def('EchoTime', echotime)
    meta.set_standard_def('RepetitionTime', repetition_time)
    meta.set_standard_def('Manufacturer', 'Varian')
    meta.set_standard_def('ProtocolName', dic['procpar']['seqfil']['values'][0])
    meta.set_standard_def('PatientName', dic['procpar']['comment']['values'][0])

    # stuff that is nice to have:
    try:
        meta.set_standard_def('SoftwareVersions', dic['procpar']['parver']['values'][0])
        meta.set_standard_def('TxCoil', dic['procpar']['rfcoil']['values'][0])
        meta.set_standard_def('RxCoil', dic['procpar']['rfcoil']['values'][0])
    except KeyError:
        warnings.warn('Expected standard metadata keying failed')
    try:
        meta.set_standard_def('InversionTime', float(dic['procpar']['ti']['values'][0]))
    except KeyError:
        pass
    try:
        meta.set_standard_def('ExcitationFlipAngle', float(dic['procpar']['flip1']['values'][0]))
    except KeyError:
        pass
    conversion_time = datetime.now().isoformat(sep='T', timespec='milliseconds')
    meta.set_standard_def('ConversionTime', conversion_time)
    meta.set_standard_def('OriginalFile', [basename(args.file)])

    # k-space
    meta.set_standard_def('kSpace', [False, False, False])

    # set tag dimensions
    meta.set_dim_info(0, "DIM_COIL")
    meta.set_dim_info(1, args.tag6)

    # Stuff full headers into user fields
    if args.dump_headers:
        meta.set_user_def(key='VarianProcpar',
                          doc='Varian procpar metadata.',
                          value=dic['procpar'])

    # File names
    if args.fileout:
        fname_out = [args.fileout, ]
    else:
        fname_out = [splitext(basename(args.file))[0], ]

    return [gen_nifti_mrs_hdr_ext(data, dwelltime, meta, orientation.Q44, no_conj=True), ], fname_out


def _varian_orientation_1d(params):
    '''Calculate 1d slice orientation from parameters struct
     extract voxel positions if available
     NB:  0d spect -- thk etc not defined
          1d slice -- thk, and then pss along the usual (theta, phi, psi) angles (and orient)
          SVS voxel - vorient, vpsi, vphi, vtheta: define angle of voxel;
                    - pos1, pos2, pos3           : define position of voxel
                    - vox1, vox2, vox3, thkunit  : define size of voxel (no thkunit is mm)
    '''
    warnings.warn('Not yet implemented')


def _varian_orientation_3d(params):
    '''Calculate single voxel spectroscopy orientation from parameters struct'''

    angle_lr = params['procpar']['vpsi']['values'][0]
    angle_ap = params['procpar']['vtheta']['values'][0]
    angle_hf = params['procpar']['vphi']['values'][0]
    angles = [angle_lr, angle_ap, -angle_hf]

    dim_lr = params['procpar']['vox1']['values'][0]
    dim_ap = params['procpar']['vox2']['values'][0]
    dim_hf = params['procpar']['vox3']['values'][0]
    dimensions = [dim_lr, dim_ap, dim_hf] * 1e-2

    shift_lr = params['procpar']['pos1']['values'][0]
    shift_ap = params['procpar']['pos2']['values'][0]
    shift_hf = params['procpar']['pos3']['values'][0]
    shift = [-shift_lr, -shift_ap, shift_hf] * 1e-2

    return calc_affine(angles, dimensions, shift)



'''Simple test for SpectralWidth compliance'''

import subprocess
from pathlib import Path
import json

import numpy as np

from .io_for_tests import read_nifti_mrs

# Data paths
siemens_path = Path(__file__).parent / 'spec2nii_test_data' / 'Siemens'
vb_path = siemens_path / 'VBData' / 'Twix/meas_MID151_svs_se_C_T15_S10_10_FID108741.dat'


def test_insertion_spectralwidth(tmp_path):
    subprocess.check_call(['spec2nii', 'twix',
                           '-e', 'image',
                           '-f', 'vb',
                           '-o', tmp_path,
                           '-j', str(vb_path)])

    img_t = read_nifti_mrs(tmp_path / 'vb.nii.gz')

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert 'SpectralWidth' in hdr_ext
    assert np.isclose(hdr_ext['SpectralWidth'], 1 / img_t.header['pixdim'][4])



""" spec2nii module containing functions specific to Siemens TWIX format
Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2020 University of Oxford
"""
import re
from os.path import basename
from datetime import datetime

import numpy as np
from nifti_mrs.create_nmrs import gen_nifti_mrs_hdr_ext
from nifti_mrs.hdr_ext import Hdr_Ext

import spec2nii.GSL.gslfunctions as GSL
from spec2nii.dcm2niiOrientation.orientationFuncs import dcm_to_nifti_orientation
from spec2nii import __version__ as spec2nii_ver


# Define some default dimension information.
# Default spatial x,y,z is Lin, Phs, Seg in VB but in VE it is Lin, Par, Seg
# In VB set is used as the repetition direction, whilst in VE Ave is used.
# However it appears some CMRR sequences on VE still use the VB.
defaults = {'vb': {'Col': 'time',
                   'Lin': 'x',
                   'Phs': 'y',
                   'Seg': 'z',
                   'Cha': 'DIM_COIL',
                   'Set': 'DIM_DYN',
                   'Rep': 'DIM_DYN'},
            'vd': {'Col': 'time',
                   'Lin': 'x',
                   'Par': 'y',
                   'Seg': 'z',
                   'Cha': 'DIM_COIL',
                   'Ave': 'DIM_DYN',
                   'Set': 'DIM_DYN',
                   'Rep': 'DIM_DYN',
                   'Eco': 'DIM_EDIT'}}

xa_product_seq = ['svs_se', 'svs_st', 'svs_slaser']


class IncompatibleSoftwareVersion(Exception):
    pass


def xa_or_vx(hdr):
    """Determine siemens software type

    :param hdr: mapVBVDHdr object
    :type img: mapvbvd.read_twix_hdr.twix_hdr
    :raises IncompatibleSoftwareVersion: Raised if not a VX or XA twix file
    :return: String either 'xa' or 'vx'
    :rtype: str
    """
    if re.search(r'syngo MR XA\d{2}', hdr['Dicom']['SoftwareVersions']):
        return 'xa'
    elif re.search(r'syngo MR [A-Z]\d{2}', hdr['Dicom']['SoftwareVersions']):
        return 'vx'
    else:
        raise IncompatibleSoftwareVersion(
            'spec2nii does not recognise this Siemens software version'
            f' ({hdr["Dicom"]["SoftwareVersions"]}).'
            ' spec2nii is tested on VA-VE, and XA20 and XA30 DICOM files.')


def seq_name(hdr):
    """Return sequence (binary) name

    :param hdr: twixobject hdr
    :type hdr: dict
    :return: Sequence name string
    :rtype: str
    """
    return hdr['Meas'][('tSequenceString')]


def is_xa_product(hdr):
    """If the baseline is xa and the sequence is a svs product sequence return True."""
    def is_product():
        if any(seq_name(hdr) == seq for seq in xa_product_seq):
            return True

    return xa_or_vx(hdr) == 'xa' and is_product()


def id_sequence_type(twixhdr):
    if seq_name(twixhdr) == 'fid':
        n_voxels = 0
    elif twixhdr.Meas.lFinalMatrixSizePhase \
            and twixhdr.Meas.lFinalMatrixSizeRead:
        n_voxels = twixhdr.Meas.lFinalMatrixSizeSlice \
            * twixhdr.Meas.lFinalMatrixSizePhase \
            * twixhdr.Meas.lFinalMatrixSizeRead
    elif twixhdr.Meas.lFinalMatrixSizeSlice:
        n_voxels = twixhdr.Meas.lFinalMatrixSizeSlice
    else:
        # If lFinalMatrixSize{Slice,Phase,Read} are all empty
        # Either unlocalised or unusually filled in headers.
        # Assume 1 voxel for either SVS or unlocalised case.
        # RM's SPECIAL sequence hits this. See https://github.com/wexeee/spec2nii/issues/6.
        n_voxels = 1

    if n_voxels > 1:
        return 'mrsi'
    elif n_voxels == 0:
        return 'fid'
    elif n_voxels == 1:
        return 'svs'


def process_twix(twixObj, base_name_out, name_in, dataKey, dim_overrides, quiet=False, verbose=False, remove_os=False):
    """Process a twix file. Identify type of MRS and then pass to the relevant function."""

    if id_sequence_type(twixObj.hdr) == 'mrsi':
        return process_mrsi(twixObj, base_name_out, name_in, dataKey, quiet=quiet, verbose=verbose)
    elif id_sequence_type(twixObj.hdr) == 'fid':
        return process_fid(
            twixObj,
            base_name_out,
            name_in,
            dataKey,
            dim_overrides,
            remove_os,
            quiet=quiet,
            verbose=verbose)
    elif id_sequence_type(twixObj.hdr) == 'svs':
        return process_svs(
            twixObj,
            base_name_out,
            name_in,
            dataKey,
            dim_overrides,
            remove_os,
            quiet=quiet,
            verbose=verbose)


def process_mrsi(twixObj, base_name_out, name_in, dataKey, remove_os=True, quiet=False, verbose=False):
    """Identify correct MRSI pathway, either simple internal reconstruction or to ismrmrd"""

    if not quiet:
        print('Generating empty NIfTI-MRS file for MRSI input.')
    if verbose:
        print('Spec2nii is not a reconstruction program.')
        print(
            'Insufficient information is stored in the twix/.dat header to carry out '
            'generalised reconstruction of arbitrary k,t-space data.')
        print(
            'This routine generates an empty file, '
            'so that data reconstructed through another '
            'pathway can be inserted into it.')

    # Data size
    data_size = []
    if twixObj.hdr.Meas.lFinalMatrixSizePhase:
        data_size.append(int(twixObj.hdr.Meas.lFinalMatrixSizePhase))
    else:
        data_size.append(int(1))
    if twixObj.hdr.Meas.lFinalMatrixSizeRead:
        data_size.append(int(twixObj.hdr.Meas.lFinalMatrixSizeRead))
    else:
        data_size.append(int(1))
    if twixObj.hdr.Meas.lFinalMatrixSizeSlice:
        data_size.append(int(twixObj.hdr.Meas.lFinalMatrixSizeSlice))
    else:
        data_size.append(int(1))

    data_size.append(int(twixObj.hdr.Meas.lVectorSize))

    # Perform Orientation calculations
    # 1) Calculate dicom like imageOrientationPatient, imagePositionPatient, pixelSpacing and slicethickness
    orient = twix2DCMOrientation(twixObj['hdr'], verbose=verbose)
    imageOrientationPatient, imagePositionPatient, pixelSpacing, slicethickness, dim_swapped = orient

    # 2) In the style of dcm2niix calculate the affine matrix
    orientation = dcm_to_nifti_orientation(imageOrientationPatient,
                                           imagePositionPatient,
                                           np.append(pixelSpacing, slicethickness),
                                           data_size[:3],
                                           half_shift=True,
                                           verbose=verbose)

    # Account for the first two dimensions being swapped if main orientation is Transverse
    # See CSIOrientations function for logic
    if dim_swapped:
        data_size = [data_size[1], data_size[0]] + data_size[2:]
    data = np.zeros(data_size, dtype=complex)

    # Extract dwellTime
    dwellTime = twixObj['hdr']['MeasYaps'][('sRXSPEC', 'alDwellTime', '0')] / 1E9
    if remove_os:
        dwellTime *= 2

    # Extract metadata
    meta_obj = extractTwixMetadata(twixObj['hdr'], basename(twixObj[dataKey].filename))

    # Identify what those indices are
    # If cha is one: loop over 3rd and higher dims and make 2D images
    # If cha isn't present one: loop over 2nd and higher dims and make 1D images
    # Don't write here, just fill up class property lists for later writing
    if base_name_out:
        out_str = base_name_out
    else:
        out_str = name_in.split('.')[0]

    out_str += '_empty'

    return [gen_nifti_mrs_hdr_ext(data, dwellTime, meta_obj, orientation.Q44), ], [out_str, ]


def process_svs(twixObj, base_name_out, name_in, dataKey, dim_overrides, remove_os, quiet=False, verbose=False):
    """Process a twix file into a NIfTI MRS file.
    Inputs:
        twixObj: object from mapVBVD.
        base_name_out: Core string of output file.
        name_in: name of input file.
        dataKey: eval info flag name,
        remove_os: Remove time-doain overrides
        quiet: True to suppress text output.
    """

    # Set squeeze data
    twixObj[dataKey].flagRemoveOS = remove_os
    twixObj[dataKey].squeeze = True
    squeezedData = twixObj[dataKey]['']

    if not quiet:
        print(f'Found data of size {squeezedData.shape}.')

    # Conjugate the data from the twix file to match the phase conventions of the format
    squeezedData = squeezedData.conj()

    # Perform Orientation calculations
    # 1) Calculate dicom like imageOrientationPatient,imagePositionPatient,pixelSpacing and slicethickness
    orient = twix2DCMOrientation(twixObj['hdr'], force_svs=True, verbose=verbose)
    imageOrientationPatient, imagePositionPatient, pixelSpacing, slicethickness, _ = orient

    # 2) In the style of dcm2niix calculate the affine matrix
    orientation = dcm_to_nifti_orientation(imageOrientationPatient,
                                           imagePositionPatient,
                                           np.append(pixelSpacing, slicethickness),
                                           (1, 1, 1),
                                           verbose=verbose)

    # # 2) in style of dcm2niix
    # #   a) calculate Q44
    # xyzMM = np.append(pixelSpacing, slicethickness)
    # Q44 = nifti_dicom2mat(imageOrientationPatient, imagePositionPatient, xyzMM, verbose=verbose)

    # #   b) calculate nifti quaternion parameters
    # Q44[:2, :] *= -1

    # # 3) place in data class for nifti orientation parameters
    # orientation = NIFTIOrient(Q44)

    # Extract dwellTime
    dwellTime = twixObj['hdr']['MeasYaps'][('sRXSPEC', 'alDwellTime', '0')] / 1E9
    if remove_os:
        dwellTime *= 2

    # Extract metadata
    meta_obj = extractTwixMetadata(twixObj['hdr'], basename(twixObj[dataKey].filename))

    # Identify what those indices are
    # If cha is one: loop over 3rd and higher dims and make 2D images
    # If cha isn't present one: loop over 2nd and higher dims and make 1D images
    # Don't write here, just fill up class property lists for later writing
    if base_name_out:
        mainStr = base_name_out
    else:
        mainStr = name_in.split('.')[0]

    dims = twixObj[dataKey].sqzDims
    if dims[0] != 'Col':
        # This is very unlikely to occur but would cause complete failure.
        raise ValueError('Col is expected to be the first dimension in the Twix file, it is not.')

    curr_defaults = defaults[twixObj[dataKey].softwareVersion]
    dim_order = twixObj[dataKey].sqzDims[1:]

    # SPECIAL CASE FOR XA 20/30 product sequences.
    # A single? reference scan is encoded in the first element of the phs loop
    # If so strip out the scan and add it as a separate output file.
    if is_xa_product(twixObj['hdr'])\
            and 'Phs' in dim_order:

        phs_dim = dim_order.index('Phs') + 1
        ref_index = [slice(None), ] * squeezedData.ndim
        ref_index[phs_dim] = slice(1)
        xa_ref_scans = squeezedData[tuple(ref_index)]
        ref_index[phs_dim] = slice(1, None, 1)
        squeezedData_noref = squeezedData[tuple(ref_index)]

        if squeezedData_noref.shape[phs_dim] == 1:
            dim_order.remove('Phs')
            squeezedData = squeezedData_noref.squeeze()

        # To do, handle any other dimensions properly
        xa_ref_scans = xa_ref_scans.squeeze()
        xa_zero_index = xa_ref_scans[0, :, :] == 0.0
        xa_ref_scans = xa_ref_scans[:, ~xa_zero_index]\
            .reshape(xa_ref_scans.shape[0], xa_ref_scans.shape[1], -1)
    else:
        xa_ref_scans = None

    # AG 03/22/2023 Moved the NIFTI/Out lists and if statement up to work with new Hyper function below.
    # AG 03/22/2023 Create Lists for assembled data - Starts with XA reference..
    nifti_mrs_out = []
    filename_out  = []

    if xa_ref_scans is not None:
        # Pad with three singleton dimensions (x,y,z)
        newshape = (1, 1, 1) + xa_ref_scans.shape
        if xa_ref_scans.ndim > 2:
            ref_tags = ['DIM_COIL', 'DIM_DYN', None]
        else:
            ref_tags = ['DIM_COIL', None, None]

        meta_obj_ref = extractTwixMetadata(twixObj['hdr'], basename(twixObj[dataKey].filename))
        meta_obj_ref.set_standard_def('WaterSuppressed', True)

        nifti_mrs_out.append(
            assemble_nifti_mrs(
                xa_ref_scans.reshape(newshape),
                dwellTime,
                orientation,
                meta_obj_ref,
                dim_tags=ref_tags))

        filename_out.append(mainStr + '_ref')

    # Make list of tags (both default and user specified)
    dim_tags = []
    unknown_counter = 0
    for do in dim_order:
        if do in curr_defaults.keys():
            dim_tags.append(curr_defaults[do])
        else:
            dim_tags.append(f'DIM_USER_{unknown_counter}')
            unknown_counter += 1

    # Now process the user specified order
    for dim_index in range(3):
        if dim_overrides['dims'][dim_index]:
            if dim_overrides['dims'][dim_index] in dim_order:
                curr_index = dim_order.index(dim_overrides['dims'][dim_index])
                dim_order[dim_index], dim_order[curr_index] = dim_order[curr_index], dim_order[dim_index]
                dim_tags[dim_index], dim_tags[curr_index] = dim_tags[curr_index], dim_tags[dim_index]
            else:
                dim_order.insert(dim_index, dim_overrides['dims'][0])
                if dim_overrides['dims'][dim_index] in curr_defaults.keys():
                    dim_tags.insert(dim_index, curr_defaults['tags'][dim_overrides['dims'][dim_index]])
                else:
                    dim_tags.insert(dim_index, f'DIM_USER_{unknown_counter}')
                    unknown_counter += 1

    # Override with any of the specified tags
    for idx, tag in enumerate(dim_overrides['tags']):
        if tag:
            dim_tags[idx] = tag

    # Permute the order of dimension in the data
    original = list(range(1, squeezedData.ndim))
    new = [twixObj[dataKey].sqzDims.index(dd) for dd in dim_order]
    reord_data = np.moveaxis(squeezedData, original, new)

    # Special-cased sequences
    # MGS SVS: MEGA-PRESS, HERMES, etc.
    if twixObj['hdr']['Meas'][('tSequenceString')] in ('mgs_svs_ed', ):                     # MGS SVS
        from spec2nii.Siemens.twix_special_case import mgs_svs_ed_twix
        reord_data, meta_obj, dim_tags = mgs_svs_ed_twix(twixObj,
                                                         reord_data,
                                                         meta_obj,
                                                         dim_tags)

    # AG 03/22/2023 - New Hyper Function
    elif 'smm_svs_herc_hyper' in twixObj['hdr']['Meas'][('tSequenceFileName')]:
        from spec2nii.Siemens.twix_special_case import smm_svs_herc_hyper                   # Hyper Data

        hyp_names  = ['35ms Water Reference',
                      '80ms Water Reference',
                      '35ms PRESS',
                      '80ms HERCULES']                                                      # Hyper Subscan Names
        hyp_suffix = ['ref_short_te', 'ref_edited', 'short_te', 'edited']                   # Hyper Suffixes

        # Reording the Data & Adjusting the Header Appropriately
        for ii in range(len(hyp_names)):                                                    # Iterate over Subscans
            print(f'{ii:3d} {hyp_names[ii]:<20}', end='\r')
            reord_data_, meta_obj_, dim_tags_ = smm_svs_herc_hyper(twixObj,
                                                                   reord_data,
                                                                   meta_obj,
                                                                   dim_tags,
                                                                   subseq=ii,
                                                                   subseq_name=hyp_names[ii])

            if reord_data_.ndim <= 4:                                                       # 4 or less Dims
                newshape  = (1, 1, 1) + reord_data_.shape                                   # Pad 3 singleton dims
                nifti_mrs_out.append(assemble_nifti_mrs(reord_data_.reshape(newshape),
                                                        dwellTime,
                                                        orientation,
                                                        meta_obj_))
                filename_out.append(f'{mainStr}_{hyp_suffix[ii]}')

        return nifti_mrs_out, filename_out

    # HERCULES Data
    elif (xa_or_vx(twixObj['hdr']) == 'xa' and 'smm_svs_herc' in twixObj['hdr']['Meas'][('tSequenceFileName')]):
        from spec2nii.Siemens.twix_special_case import mgs_svs_ed_twix
        reord_data, meta_obj, dim_tags = mgs_svs_ed_twix(twixObj, reord_data, meta_obj, dim_tags)

    elif re.search(
            r'svs_slaser(voi)?_dkd',
            twixObj['hdr']['Meas'][('tSequenceFileName')],
            re.IGNORECASE):
        from spec2nii.Siemens.twix_special_case import slaser_dkd
        for data, meta, name in slaser_dkd(twixObj, reord_data, meta_obj, dim_tags):

            nifti_mrs_out.append(
                gen_nifti_mrs_hdr_ext(
                    data,
                    dwellTime,
                    meta,
                    orientation.Q44,
                    no_conj=True))

            filename_out.append(f'{mainStr}{name}')

        return nifti_mrs_out, filename_out
    else:
        # Set dim tags in meta now as no additional info
        for idx, dt in enumerate(dim_tags):
            meta_obj.set_dim_info(idx, dt)

    if reord_data.ndim <= 4:
        # Pad with three singleton dimensions (x,y,z)
        newshape = (1, 1, 1) + reord_data.shape

        nifti_mrs_out.append(assemble_nifti_mrs(reord_data.reshape(newshape),
                                                dwellTime,
                                                orientation,
                                                meta_obj))

        filename_out.append(mainStr)

    else:
        # loop over any dimensions over 4
        for index in np.ndindex(reord_data.shape[4:]):
            modIndex = (slice(None), slice(None), slice(None), slice(None)) + index

            # Pad with three singleton dimensions (x,y,z)
            newshape = (1, 1, 1) + reord_data[modIndex].shape

            nifti_mrs_out.append(
                assemble_nifti_mrs(reord_data[modIndex].reshape(newshape),
                                   dwellTime,
                                   orientation,
                                   meta_obj))

            # Create strings
            out_name = f'{mainStr}'
            for idx, ii in enumerate(index):
                indexStr = dim_order[3 + idx]
                out_name += f'_{indexStr}{ii: 03d}'

            filename_out.append(out_name)

    return nifti_mrs_out, filename_out


def process_fid(twixObj, base_name_out, name_in, dataKey, dim_overrides, remove_os, quiet=False, verbose=False):
    """Process a twix file into a NIfTI MRS file.
    Inputs:
        twixObj: object from mapVBVD.
        base_name_out: Core string of output file.
        name_in: name of input file.
        dataKey: eval info flag name,
        remove_os: Remove time-doain overrides
        quiet: True to suppress text output.
    """

    # Set squeeze data
    twixObj[dataKey].flagRemoveOS = remove_os
    twixObj[dataKey].squeeze = True
    squeezedData = twixObj[dataKey]['']

    if not quiet:
        print(f'Found data of size {squeezedData.shape}.')

    # Conjugate the data from the twix file to match the phase conventions of the format
    squeezedData = squeezedData.conj()

    # Perform Orientation calculations
    # 1) Calculate dicom like imageOrientationPatient,imagePositionPatient,pixelSpacing and slicethickness
    orient = twix2DCMOrientation(twixObj['hdr'], force_svs=True, verbose=verbose)
    imageOrientationPatient, imagePositionPatient, pixelSpacing, slicethickness, _ = orient

    # 2) In the style of dcm2niix calculate the affine matrix
    orientation = dcm_to_nifti_orientation(imageOrientationPatient,
                                           imagePositionPatient,
                                           np.append(pixelSpacing, slicethickness),
                                           (1, 1, 1),
                                           verbose=verbose)

    # # 2) in style of dcm2niix
    # #   a) calculate Q44
    # xyzMM = np.append(pixelSpacing, slicethickness)
    # Q44 = nifti_dicom2mat(imageOrientationPatient, imagePositionPatient, xyzMM, verbose=verbose)

    # #   b) calculate nifti quaternion parameters
    # Q44[:2, :] *= -1

    # # 3) place in data class for nifti orientation parameters
    # orientation = NIFTIOrient(Q44)

    # Extract dwellTime
    dwellTime = twixObj['hdr']['MeasYaps'][('sRXSPEC', 'alDwellTime', '0')] / 1E9
    if remove_os:
        dwellTime *= 2

    # Extract metadata
    meta_obj = extractTwixMetadata(twixObj['hdr'], basename(twixObj[dataKey].filename))

    # Identify what those indices are
    # If cha is one: loop over 3rd and higher dims and make 2D images
    # If cha isn't present one: loop over 2nd and higher dims and make 1D images
    # Don't write here, just fill up class property lists for later writing
    if base_name_out:
        mainStr = base_name_out
    else:
        mainStr = name_in.split('.')[0]

    dims = twixObj[dataKey].sqzDims
    if dims[0] != 'Col':
        # This is very unlikely to occur but would cause complete failure.
        raise ValueError('Col is expected to be the first dimension in the Twix file, it is not.')

    curr_defaults = defaults[twixObj[dataKey].softwareVersion]
    dim_order = twixObj[dataKey].sqzDims[1:]

    # SPECIAL CASE FOR XA 20/30 product sequences.
    # A single? reference scan is encoded in the first element of the phs loop
    # If so strip out the scan and add it as a separate output file.
    if is_xa_product(twixObj['hdr'])\
            and 'Phs' in dim_order:

        phs_dim = dim_order.index('Phs') + 1
        ref_index = [slice(None), ] * squeezedData.ndim
        ref_index[phs_dim] = slice(1)
        xa_ref_scans = squeezedData[tuple(ref_index)]
        ref_index[phs_dim] = slice(1, None, 1)
        squeezedData_noref = squeezedData[tuple(ref_index)]

        if squeezedData_noref.shape[phs_dim] == 1:
            dim_order.remove('Phs')
            squeezedData = squeezedData_noref.squeeze()

        # To do, handle any other dimensions properly
        xa_ref_scans = xa_ref_scans.squeeze()
        xa_zero_index = xa_ref_scans[0, :, :] == 0.0
        xa_ref_scans = xa_ref_scans[:, ~xa_zero_index]\
            .reshape(xa_ref_scans.shape[0], xa_ref_scans.shape[1], -1)
    else:
        xa_ref_scans = None

    # Make list of tags (both default and user specified)
    dim_tags = []
    unknown_counter = 0
    for do in dim_order:
        if do in curr_defaults.keys():
            dim_tags.append(curr_defaults[do])
        else:
            dim_tags.append(f'DIM_USER_{unknown_counter}')
            unknown_counter += 1

    # Now process the user specified order
    for dim_index in range(3):
        if dim_overrides['dims'][dim_index]:
            if dim_overrides['dims'][dim_index] in dim_order:
                curr_index = dim_order.index(dim_overrides['dims'][dim_index])
                dim_order[dim_index], dim_order[curr_index] = dim_order[curr_index], dim_order[dim_index]
                dim_tags[dim_index], dim_tags[curr_index] = dim_tags[curr_index], dim_tags[dim_index]
            else:
                dim_order.insert(dim_index, dim_overrides['dims'][0])
                if dim_overrides['dims'][dim_index] in curr_defaults.keys():
                    dim_tags.insert(dim_index, curr_defaults['tags'][dim_overrides['dims'][dim_index]])
                else:
                    dim_tags.insert(dim_index, f'DIM_USER_{unknown_counter}')
                    unknown_counter += 1

    # Override with any of the specified tags
    for idx, tag in enumerate(dim_overrides['tags']):
        if tag:
            dim_tags[idx] = tag

    # Permute the order of dimension in the data
    original = list(range(1, squeezedData.ndim))
    new = [twixObj[dataKey].sqzDims.index(dd) for dd in dim_order]
    reord_data = np.moveaxis(squeezedData, original, new)

    # Now assemble data
    nifti_mrs_out = []
    filename_out = []
    if reord_data.ndim <= 4:
        # Pad with three singleton dimensions (x,y,z)
        newshape = (1, 1, 1) + reord_data.shape

        nifti_mrs_out.append(assemble_nifti_mrs(reord_data.reshape(newshape),
                                                dwellTime,
                                                orientation,
                                                meta_obj,
                                                dim_tags))

        filename_out.append(mainStr)

    else:
        # loop over any dimensions over 4
        for index in np.ndindex(reord_data.shape[4:]):
            modIndex = (slice(None), slice(None), slice(None), slice(None)) + index

            # Pad with three singleton dimensions (x,y,z)
            newshape = (1, 1, 1) + reord_data[modIndex].shape

            nifti_mrs_out.append(
                assemble_nifti_mrs(reord_data[modIndex].reshape(newshape),
                                   dwellTime,
                                   orientation,
                                   meta_obj,
                                   dim_tags))

            # Create strings
            out_name = f'{mainStr}'
            for idx, ii in enumerate(index):
                indexStr = dim_order[3 + idx]
                out_name += f'_{indexStr}{ii: 03d}'

            filename_out.append(out_name)

    if xa_ref_scans is not None:
        # Pad with three singleton dimensions (x,y,z)
        newshape = (1, 1, 1) + xa_ref_scans.shape
        if xa_ref_scans.ndim > 2:
            ref_tags = ['DIM_COIL', 'DIM_DYN', None]
        else:
            ref_tags = ['DIM_COIL', None, None]

        meta_obj_ref = extractTwixMetadata(twixObj['hdr'], basename(twixObj[dataKey].filename))
        meta_obj_ref.set_standard_def('WaterSuppressed', True)

        nifti_mrs_out.append(
            assemble_nifti_mrs(
                xa_ref_scans.reshape(newshape),
                dwellTime,
                orientation,
                meta_obj_ref,
                ref_tags))

        filename_out.append(mainStr + '_ref')

    return nifti_mrs_out, filename_out


def assemble_nifti_mrs(data, dwellTime, orientation, meta_obj, dim_tags=None):

    if dim_tags is not None:
        for idx, dt in zip(range(data.ndim - 4), dim_tags):
            meta_obj.set_dim_info(idx, dt)

    return gen_nifti_mrs_hdr_ext(data, dwellTime, meta_obj, orientation.Q44, no_conj=True)


def twix2DCMOrientation(mapVBVDHdr, force_svs=False, verbose=False):
    """ Convert twix orientation information to DICOM equivalent.

    Convert orientation to DICOM imageOrientationPatient, imagePositionPatient,
    pixelSpacing and sliceThickness field values.

    Args:
        mapVBVDHdr (dict): Header info interpreted by pymapVBVD
        force_svs (bool,optionl): Forces svs orientation information (size) to be read
        verbose (bool,optionl)
    Returns:
        imageOrientationPatient
        imagePositionPatient
        pixelSpacing
        sliceThickness

    """
    # Only single-slices are supported -- throw an error otherwise

    if ('sGroupArray', 'asGroup', '0', 'nSize') in mapVBVDHdr['MeasYaps']:
        nSlices = mapVBVDHdr['MeasYaps'][('sGroupArray', 'asGroup', '0', 'nSize')]
        if nSlices != 1.0:
            raise ValueError('In slice-selective spectroscopy, only the first slice is supported')

    # Orientation information
    # Added the force_svs because in some sequences there are slice objects initialised
    # and recorded but this seems sporadic behaviour.
    if ('sSliceArray', 'asSlice', '0', 'sNormal', 'dSag') in mapVBVDHdr['MeasYaps']\
            and not force_svs:
        # This is for slice-selective spectroscopy
        NormaldSag = mapVBVDHdr['MeasYaps'][('sSliceArray', 'asSlice', '0', 'sNormal', 'dSag')]
    elif ('sSpecPara', 'sVoI', 'sNormal', 'dSag') in mapVBVDHdr['MeasYaps']:
        NormaldSag = mapVBVDHdr['MeasYaps'][('sSpecPara', 'sVoI', 'sNormal', 'dSag')]
    else:
        NormaldSag = 0.0

    if ('sSliceArray', 'asSlice', '0', 'sNormal', 'dCor') in mapVBVDHdr['MeasYaps']\
            and not force_svs:
        NormaldCor = mapVBVDHdr['MeasYaps'][('sSliceArray', 'asSlice', '0', 'sNormal', 'dCor')]
    elif ('sSpecPara', 'sVoI', 'sNormal', 'dCor') in mapVBVDHdr['MeasYaps']:
        NormaldCor = mapVBVDHdr['MeasYaps'][('sSpecPara', 'sVoI', 'sNormal', 'dCor')]
    else:
        NormaldCor = 0.0

    if ('sSliceArray', 'asSlice', '0', 'sNormal', 'dTra') in mapVBVDHdr['MeasYaps']\
            and not force_svs:
        NormaldTra = mapVBVDHdr['MeasYaps'][('sSliceArray', 'asSlice', '0', 'sNormal', 'dTra')]
    elif ('sSpecPara', 'sVoI', 'sNormal', 'dTra') in mapVBVDHdr['MeasYaps']:
        NormaldTra = mapVBVDHdr['MeasYaps'][('sSpecPara', 'sVoI', 'sNormal', 'dTra')]
    else:
        NormaldTra = 0.0

    if ('sSliceArray', 'asSlice', '0', 'dInPlaneRot') in mapVBVDHdr['MeasYaps']\
            and not force_svs:
        inplaneRotation = mapVBVDHdr['MeasYaps'][('sSliceArray', 'asSlice', '0', 'dInPlaneRot')]
    elif ('sSpecPara', 'sVoI', 'dInPlaneRot') in mapVBVDHdr['MeasYaps']:
        inplaneRotation = mapVBVDHdr['MeasYaps'][('sSpecPara', 'sVoI', 'dInPlaneRot')]
    else:
        inplaneRotation = 0.0

    TwixSliceNormal = np.array([NormaldSag, NormaldCor, NormaldTra], dtype=float)
    # If all zeros make a 'normal' orientation (e.g. for unlocalised data)
    if not TwixSliceNormal.any():
        TwixSliceNormal[0] += 1.0

    if ('sSliceArray', 'asSlice', '0', 'dReadoutFOV') in mapVBVDHdr['MeasYaps']\
            and not force_svs:
        RoFoV = mapVBVDHdr['MeasYaps'][('sSliceArray', 'asSlice', '0', 'dReadoutFOV')]
        PeFoV = mapVBVDHdr['MeasYaps'][('sSliceArray', 'asSlice', '0', 'dPhaseFOV')]
    elif ('sSpecPara', 'sVoI', 'dReadoutFOV') in mapVBVDHdr['MeasYaps']:
        RoFoV = mapVBVDHdr['MeasYaps'][('sSpecPara', 'sVoI', 'dReadoutFOV')]
        PeFoV = mapVBVDHdr['MeasYaps'][('sSpecPara', 'sVoI', 'dPhaseFOV')]
    else:
        RoFoV = 10000.0
        PeFoV = 10000.0

    if ('sSliceArray', 'asSlice', '0', 'dThickness') in mapVBVDHdr['MeasYaps']\
            and not force_svs:
        sliceThickness = mapVBVDHdr['MeasYaps'][('sSliceArray', 'asSlice', '0', 'dThickness')]
    elif ('sSpecPara', 'sVoI', 'dThickness') in mapVBVDHdr['MeasYaps']:
        sliceThickness = mapVBVDHdr['MeasYaps'][('sSpecPara', 'sVoI', 'dThickness')]
    else:
        sliceThickness = 10000.0

    # Position info (including table position)
    if ('sSliceArray', 'asSlice', '0', 'sPosition', 'dSag') in mapVBVDHdr['MeasYaps']\
            and not force_svs:
        PosdSag = mapVBVDHdr['MeasYaps'][('sSliceArray', 'asSlice', '0', 'sPosition', 'dSag')]
    elif ('sSpecPara', 'sVoI', 'sPosition', 'dSag') in mapVBVDHdr['MeasYaps']:
        PosdSag = mapVBVDHdr['MeasYaps'][('sSpecPara', 'sVoI', 'sPosition', 'dSag')]
    else:
        PosdSag = 0.0

    if ('sSliceArray', 'asSlice', '0', 'sPosition', 'dCor') in mapVBVDHdr['MeasYaps']\
            and not force_svs:
        PosdCor = mapVBVDHdr['MeasYaps'][('sSliceArray', 'asSlice', '0', 'sPosition', 'dCor')]
    elif ('sSpecPara', 'sVoI', 'sPosition', 'dCor') in mapVBVDHdr['MeasYaps']:
        PosdCor = mapVBVDHdr['MeasYaps'][('sSpecPara', 'sVoI', 'sPosition', 'dCor')]
    else:
        PosdCor = 0.0

    if ('sSliceArray', 'asSlice', '0', 'sPosition', 'dTra') in mapVBVDHdr['MeasYaps']\
            and not force_svs:
        PosdTra = mapVBVDHdr['MeasYaps'][('sSliceArray', 'asSlice', '0', 'sPosition', 'dTra')]
    elif ('sSpecPara', 'sVoI', 'sPosition', 'dTra') in mapVBVDHdr['MeasYaps']:
        PosdTra = mapVBVDHdr['MeasYaps'][('sSpecPara', 'sVoI', 'sPosition', 'dTra')]
    else:
        PosdTra = 0.0

    if ('lScanRegionPosSag',) in mapVBVDHdr['MeasYaps']:
        PosdSag += mapVBVDHdr['MeasYaps'][('lScanRegionPosSag',)]
    if ('lScanRegionPosCor',) in mapVBVDHdr['MeasYaps']:
        PosdCor += mapVBVDHdr['MeasYaps'][('lScanRegionPosCor',)]
    if ('lScanRegionPosTra',) in mapVBVDHdr['MeasYaps']:
        PosdTra += mapVBVDHdr['MeasYaps'][('lScanRegionPosTra',)]

    if id_sequence_type(mapVBVDHdr) == 'mrsi':
        data_size_pe = 1
        if mapVBVDHdr.Meas.lFinalMatrixSizePhase:
            data_size_pe = mapVBVDHdr.Meas.lFinalMatrixSizePhase

        data_size_ro = 1
        if mapVBVDHdr.Meas.lFinalMatrixSizeRead:
            data_size_ro = mapVBVDHdr.Meas.lFinalMatrixSizeRead

        data_size_sl = 1
        if mapVBVDHdr.Meas.lFinalMatrixSizeSlice:
            data_size_sl = mapVBVDHdr.Meas.lFinalMatrixSizeSlice

        base_pos_info = np.array([PosdSag, PosdCor, PosdTra], dtype=float)

        imageOrientationPatient, imagePositionPatient, pixelSpacing, sliceThickness, dim_swapped = CSIOrientations(
            TwixSliceNormal,
            inplaneRotation,
            PeFoV,
            RoFoV,
            sliceThickness,
            data_size_pe,
            data_size_ro,
            data_size_sl,
            base_pos_info,
            verbose)

    else:
        dColVec_vector, dRowVec_vector = GSL.calc_prs(TwixSliceNormal, inplaneRotation, verbose)
        imageOrientationPatient = np.stack((dRowVec_vector, dColVec_vector), axis=0)

        pixelSpacing = np.array([PeFoV, RoFoV])  # [RoFoV PeFoV];

        imagePositionPatient = np.array([PosdSag, PosdCor, PosdTra], dtype=float)

        dim_swapped = False

    if verbose:
        print(f'imagePositionPatient is {imagePositionPatient.ravel()}')
        print(f'imageOrientationPatient is \n{imageOrientationPatient}')
        print(f'{imageOrientationPatient.ravel()}')
        print(f'pixelSpacing is {pixelSpacing}')

    return imageOrientationPatient, imagePositionPatient, pixelSpacing, sliceThickness, dim_swapped


def CSIOrientations(slice_normal, ip_rot, fov_pe, fov_ro, fov_sl, n_pe, n_ro, n_sl, base_pos, verbose=False):
    SAGITTAL   = 0
    CORONAL    = 1
    TRANSVERSE = 2
    mo_case = GSL.class_ori(slice_normal[SAGITTAL], slice_normal[CORONAL], slice_normal[TRANSVERSE], False)
    dColVec_vector, dRowVec_vector = GSL.calc_prs(slice_normal, ip_rot, verbose)
    print('To achieve matching orientation to equivalent DICOM output:')

    if n_sl > 1:
        # 3D MRSI
        base_pos -= slice_normal * (fov_sl / 2 - fov_sl / n_sl / 2)
        fov_sl /= n_sl

    # Sagittal
    if mo_case == 0:
        print('Mirror along ROW/readout direction: VB = LIN, VE+ = SEG')
        dRowVec_vector *= -1.0
        pixelSpacing = np.array([fov_ro / n_ro, fov_pe / n_pe])
        dColVec_vector, dRowVec_vector = dRowVec_vector, dColVec_vector
        imagePositionPatient =\
            base_pos - (dRowVec_vector * fov_pe / 2) - (dColVec_vector * fov_ro / 2)
        dim_swapped = False
    # Coronal
    elif mo_case == 1:
        pixelSpacing = np.array([fov_ro / n_ro, fov_pe / n_pe])
        dColVec_vector, dRowVec_vector = dRowVec_vector, dColVec_vector
        imagePositionPatient =\
            base_pos - (dRowVec_vector * fov_pe / 2) - (dColVec_vector * fov_ro / 2)
        dim_swapped = False
    # Transverse
    elif mo_case == 2:
        print('Mirror along ROW/readout direction: VB = LIN, VE+ = SEG')
        dRowVec_vector *= -1.0
        pixelSpacing = np.array([fov_pe / n_pe, fov_ro / n_ro])
        print('Swap 1st and 2nd diemensions. For VB swap LIN and PHS, for VE+ swap SEG and LIN dimensions')
        dim_swapped = True
        imagePositionPatient =\
            base_pos - (dRowVec_vector * fov_ro / 2) - (dColVec_vector * fov_pe / 2)
    imageOrientationPatient = np.stack((dRowVec_vector, dColVec_vector), axis=0)

    return imageOrientationPatient, imagePositionPatient, pixelSpacing, fov_sl, dim_swapped


def examineTwix(twixObj, fileName, mraid):
    """ Print formatted twix contents"""

    print(f'Contents of file {fileName}:')

    if isinstance(twixObj, list):
        print(f'Multiraid file, {len(twixObj)} files found.')
        print(f'Selecting file {mraid}. Use -m option to change.')
        twixObj = twixObj[mraid - 1]

    evalInfoFlags = twixObj.keys()
    evalInfoFlags = [i for i in evalInfoFlags if i != 'hdr']

    print('The file contains these evalinfo flags with dimensions and sizes as follows:')
    for ev in evalInfoFlags:
        twixObj[ev].flagRemoveOS = False
        twixObj[ev].squeeze = True
        tmpSqzSize = twixObj[ev].sqzSize
        tmpSqzDims = ', '.join(twixObj[ev].sqzDims)
        print(f'{ev: <15}:\t{tmpSqzDims: <20}\t{tmpSqzSize}')


def extractTwixMetadata(mapVBVDHdr, original_file):
    """Pass to appropriate extractTwixMetadata function for software version."""
    if xa_or_vx(mapVBVDHdr) == 'vx':
        try:
            return extractTwixMetadata_vx(mapVBVDHdr, original_file)
        except KeyError as original_err:
            try:
                print(f'VX-line scanner headers caused KeyError {original_err}.')
                print('Trying XA interpreter. ')
                return extractTwixMetadata_xa(mapVBVDHdr, original_file)
            except Exception as backup_err:
                print(f'XA interpreter failed with reason {backup_err}. Raising original error.')
                raise original_err

    elif xa_or_vx(mapVBVDHdr) == 'xa':
        return extractTwixMetadata_xa(mapVBVDHdr, original_file)


def extractTwixMetadata_xa(mapVBVDHdr, original_file):
    """ Extract information from the pymapVBVD header to insert into the json sidecar.

    Args:
        dcmdata (dict): Twix headers
    Returns:
        obj (hdr_ext): NIfTI MRS hdr ext object.
    """

    # Extract required metadata and create hdr_ext object
    obj = Hdr_Ext(
        mapVBVDHdr['Dicom'][('lFrequency')] / 1E6,
        mapVBVDHdr['MeasYaps'][('sTXSPEC', 'asNucleusInfo', '0', 'tNucleus')].strip('"'))

    # Standard defined metadata
    # # 5.1 MRS specific Tags
    # 'EchoTime'
    obj.set_standard_def('EchoTime', mapVBVDHdr['Phoenix'][('alTE', '0')] * 1E-6)
    # 'RepetitionTime'
    tr = mapVBVDHdr['MeasYaps'][('alTR', '0')] / 1E6
    obj.set_standard_def('RepetitionTime', float(tr))
    # 'InversionTime'
    if ('alTI', '0') in mapVBVDHdr['MeasYaps']:
        obj.set_standard_def('InversionTime', float(mapVBVDHdr['MeasYaps'][('alTI', '0')]))
    # 'MixingTime'
    # 'ExcitationFlipAngle'
    obj.set_standard_def('ExcitationFlipAngle', float(mapVBVDHdr['MeasYaps'][('adFlipAngleDegree', '0')]))
    # 'TxOffset'
    obj.set_standard_def(
        'TxOffset',
        empty_str_or_val_to_0float(mapVBVDHdr['MeasYaps'], ('sSpecPara', 'dDeltaFrequency')))
    # 'VOI'
    # 'WaterSuppressed'
    # TO DO
    # 'WaterSuppressionType'
    # 'SequenceTriggered'
    # # 5.2 Scanner information
    # 'Manufacturer'
    obj.set_standard_def('Manufacturer', mapVBVDHdr['Dicom'][('Manufacturer')])
    # 'ManufacturersModelName'
    obj.set_standard_def('ManufacturersModelName', mapVBVDHdr['Dicom'][('ManufacturersModelName')])
    # 'DeviceSerialNumber'
    obj.set_standard_def('DeviceSerialNumber', str(mapVBVDHdr['Dicom'][('DeviceSerialNumber')]))
    # 'SoftwareVersions'
    obj.set_standard_def('SoftwareVersions', mapVBVDHdr['Dicom'][('SoftwareVersions')])
    # 'InstitutionName'
    obj.set_standard_def('InstitutionName', mapVBVDHdr['Dicom'][('InstitutionName')])
    # 'InstitutionAddress'
    obj.set_standard_def('InstitutionAddress', mapVBVDHdr['Dicom'][('InstitutionAddress')])
    # 'TxCoil'
    tx_coil_tuple = ('sCoilSelectMeas', 'aTxCoilSelectData', '0', 'asList', '0', 'sCoilElementID', 'tCoilID')
    obj.set_standard_def('TxCoil', mapVBVDHdr['MeasYaps'][tx_coil_tuple])
    # 'RxCoil'
    rx_coil_1 = ('sCoilSelectMeas', 'aRxCoilSelectData', '0', 'asList', '0', 'sCoilElementID', 'tCoilID')
    rx_coil_2 = ('asCoilSelectMeas', '0', 'asList', '0', 'sCoilElementID', 'tCoilID')
    if rx_coil_1 in mapVBVDHdr['MeasYaps']:
        obj.set_standard_def('RxCoil', mapVBVDHdr['MeasYaps'][rx_coil_1])
    elif rx_coil_2 in mapVBVDHdr['MeasYaps']:
        obj.set_standard_def('RxCoil', mapVBVDHdr['MeasYaps'][rx_coil_2])

    # # 5.3 Sequence information
    # 'SequenceName'
    obj.set_standard_def('SequenceName', mapVBVDHdr['Meas'][('tSequenceString')])
    # 'ProtocolName'
    obj.set_standard_def('ProtocolName', mapVBVDHdr['Dicom'][('tProtocolName')])
    # # 5.4 Sequence information
    # 'PatientPosition'
    obj.set_standard_def('PatientPosition', mapVBVDHdr['Meas'][('tPatientPosition')])
    # 'PatientName'
    obj.set_standard_def('PatientName', mapVBVDHdr['Meas'][('tPatientName')])
    # 'PatientID'
    # 'PatientWeight'
    try:
        obj.set_standard_def('PatientWeight', float(mapVBVDHdr['Meas'][('flUsedPatientWeight')]))
    except ValueError:
        pass
    # 'PatientDoB'
    obj.set_standard_def('PatientDoB', str(mapVBVDHdr['Meas'][('PatientBirthDay')]))
    # 'PatientSex'
    patient_sex = mapVBVDHdr['Meas'][('PatientSex')]
    if patient_sex == 1.0:
        sex_str = 'M'
    elif patient_sex == 2.0:
        sex_str = 'F'
    elif patient_sex == 3.0:
        sex_str = 'O'
    else:
        raise ValueError(f'Meas, PatientSex, should be 1, 2, or 3, but is {patient_sex}')

    obj.set_standard_def('PatientSex', sex_str)

    # # 5.5 Provenance and conversion metadata
    # 'ConversionMethod'
    obj.set_standard_def('ConversionMethod', f'spec2nii v{spec2nii_ver}')
    # 'ConversionTime'
    conversion_time = datetime.now().isoformat(sep='T', timespec='milliseconds')
    obj.set_standard_def('ConversionTime', conversion_time)
    # 'OriginalFile'
    obj.set_standard_def('OriginalFile', [original_file, ])
    # # 5.6 Spatial information
    # 'kSpace'
    obj.set_standard_def('kSpace', [False, False, False])

    # Some additional information
    obj.set_user_def(key='PulseSequenceFile',
                     value=mapVBVDHdr['Config'][('SequenceFileName')],
                     doc='Sequence binary path.')
    obj.set_user_def(key='IceProgramFile',
                     value=mapVBVDHdr['Meas'][('tICEProgramName')],
                     doc='Reconstruction binary path.')

    return obj


def extractTwixMetadata_vx(mapVBVDHdr, original_file):
    """ Extract information from the pymapVBVD header to insert into the json sidecar.

    Function Twix files from Numaris4 Vx software scanners.

    Args:
        dcmdata (dict): Twix headers
    Returns:
        obj (hdr_ext): NIfTI MRS hdr ext object.
    """

    # Extract required metadata and create hdr_ext object
    obj = Hdr_Ext(
        mapVBVDHdr['Meas'][('Frequency')] / 1E6,
        mapVBVDHdr['Meas'][('ResonantNucleus')])

    # Standard defined metadata
    # # 5.1 MRS specific Tags
    # 'EchoTime'
    obj.set_standard_def('EchoTime', mapVBVDHdr['Phoenix'][('alTE', '0')] * 1E-6)
    # 'RepetitionTime'
    if ('TR_Time') in mapVBVDHdr['Meas']:
        tr = mapVBVDHdr['Meas'][('TR_Time')] / 1E6
    else:
        tr = mapVBVDHdr['Meas'][('TR')] / 1E6
    obj.set_standard_def('RepetitionTime', float(tr))
    # 'InversionTime'
    if ('InversionTime') in mapVBVDHdr['Meas']:
        obj.set_standard_def('InversionTime', float(mapVBVDHdr['Meas'][('TI_Time')]))
    # 'MixingTime'
    # 'ExcitationFlipAngle'
    obj.set_standard_def('ExcitationFlipAngle', float(mapVBVDHdr['Meas'][('FlipAngle')]))
    # 'TxOffset'
    obj.set_standard_def('TxOffset', empty_str_or_val_to_0float(mapVBVDHdr['Meas'], ('dDeltaFrequency')))
    # 'VOI'
    # 'WaterSuppressed'
    # TO DO
    # 'WaterSuppressionType'
    # 'SequenceTriggered'
    # # 5.2 Scanner information
    # 'Manufacturer'
    obj.set_standard_def('Manufacturer', mapVBVDHdr['Dicom'][('Manufacturer')])
    # 'ManufacturersModelName'
    obj.set_standard_def('ManufacturersModelName', mapVBVDHdr['Dicom'][('ManufacturersModelName')])
    # 'DeviceSerialNumber'
    obj.set_standard_def('DeviceSerialNumber', str(mapVBVDHdr['Dicom'][('DeviceSerialNumber')]))
    # 'SoftwareVersions'
    obj.set_standard_def('SoftwareVersions', mapVBVDHdr['Dicom'][('SoftwareVersions')])
    # 'InstitutionName'
    obj.set_standard_def('InstitutionName', mapVBVDHdr['Dicom'][('InstitutionName')])
    # 'InstitutionAddress'
    obj.set_standard_def('InstitutionAddress', mapVBVDHdr['Dicom'][('InstitutionAddress')])
    # 'TxCoil'
    # 'RxCoil'
    rx_coil_1 = ('sCoilSelectMeas', 'aRxCoilSelectData', '0', 'asList', '0', 'sCoilElementID', 'tCoilID')
    rx_coil_2 = ('asCoilSelectMeas', '0', 'asList', '0', 'sCoilElementID', 'tCoilID')
    if rx_coil_1 in mapVBVDHdr['MeasYaps']:
        obj.set_standard_def('RxCoil', mapVBVDHdr['MeasYaps'][rx_coil_1])
    elif rx_coil_2 in mapVBVDHdr['MeasYaps']:
        obj.set_standard_def('RxCoil', mapVBVDHdr['MeasYaps'][rx_coil_2])
    # # 5.3 Sequence information
    # 'SequenceName'
    obj.set_standard_def('SequenceName', mapVBVDHdr['Meas'][('tSequenceString')])
    # 'ProtocolName'
    obj.set_standard_def('ProtocolName', mapVBVDHdr['Dicom'][('tProtocolName')])
    # # 5.4 Sequence information
    # 'PatientPosition'
    obj.set_standard_def('PatientPosition', mapVBVDHdr['Meas'][('PatientPosition')])
    # 'PatientName'
    obj.set_standard_def('PatientName', mapVBVDHdr['Meas'][('PatientName')])
    # 'PatientID'
    # 'PatientWeight'
    try:
        obj.set_standard_def('PatientWeight', float(mapVBVDHdr['Meas'][('flUsedPatientWeight')]))
    except ValueError:
        pass
    # 'PatientDoB'
    obj.set_standard_def('PatientDoB', str(mapVBVDHdr['Meas'][('PatientBirthDay')]))
    # 'PatientSex'
    if mapVBVDHdr['Meas'][('PatientSex')] == 1:
        sex_str = 'M'
    elif mapVBVDHdr['Meas'][('PatientSex')] == 2:
        sex_str = 'F'
    else:
        sex_str = 'O'
    obj.set_standard_def('PatientSex', sex_str)
    # # 5.5 Provenance and conversion metadata
    # 'ConversionMethod'
    obj.set_standard_def('ConversionMethod', f'spec2nii v{spec2nii_ver}')
    # 'ConversionTime'
    conversion_time = datetime.now().isoformat(sep='T', timespec='milliseconds')
    obj.set_standard_def('ConversionTime', conversion_time)
    # 'OriginalFile'
    obj.set_standard_def('OriginalFile', [original_file, ])
    # # 5.6 Spatial information
    # 'kSpace'
    obj.set_standard_def('kSpace', [False, False, False])

    # Some additional information
    obj.set_user_def(key='PulseSequenceFile',
                     value=mapVBVDHdr['Config'][('SequenceFileName')],
                     doc='Sequence binary path.')
    obj.set_user_def(key='IceProgramFile',
                     value=mapVBVDHdr['Meas'][('tICEProgramName')],
                     doc='Reconstruction binary path.')

    return obj


def empty_str_or_val_to_0float(param_dict, key):
    try:
        if param_dict[key] == '':
            return 0.0
        else:
            return param_dict[key]
    except KeyError:
        return 0.0


def _try_int(value):
    try:
        return int(value)
    except ValueError:
        return value



'''Tests for Philips DICOM format conversion.

Copyright William Clarke, University of Oxford 2021
Subject to the BSD 3-Clause License.
'''

import subprocess
from pathlib import Path
import json

import numpy as np

from .io_for_tests import read_nifti_mrs

# Data paths
philips_path = Path(__file__).parent / 'spec2nii_test_data' / 'philips'
data_path = philips_path / 'DICOM' / 'SV_phantom_H15mm' / 'IM-0020-0002-0001.dcm'
data_path_2 = philips_path / 'DICOM_enhanced_multi_dynamic' / 'svsWSAntCing_S002' / 'XX_0004'

data_path_press = philips_path / 'DICOM_enhanced_multi_dynamic' / 'press_mega' / 'XX_0006'
data_path_mpress = philips_path / 'DICOM_enhanced_multi_dynamic' / 'press_mega' / 'XX_0004'


def test_svs(tmp_path):

    subprocess.check_call(['spec2nii', 'philips_dcm',
                           '-f', 'svs',
                           '-o', tmp_path,
                           '-j',
                           str(data_path)])

    img_t = read_nifti_mrs(tmp_path / 'svs.nii.gz')

    assert img_t.shape == (1, 1, 1, 1024)
    assert np.iscomplexobj(img_t.dataobj)
    assert np.isclose(1 / img_t.header['pixdim'][4], 2000.0)

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['SpectrometerFrequency'][0] == 127.750447
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == data_path.name

    assert (tmp_path / 'svs_ref.nii.gz').is_file()


def test_svs_enhanced(tmp_path):

    subprocess.check_call(['spec2nii', 'philips_dcm',
                           '-f', 'svs',
                           '-o', tmp_path,
                           '-j',
                           str(data_path_2)])

    img_t = read_nifti_mrs(tmp_path / 'svs.nii.gz')

    assert img_t.shape == (1, 1, 1, 2048, 32)
    assert np.iscomplexobj(img_t.dataobj)
    assert np.isclose(1 / img_t.header['pixdim'][4], 2000.0)

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['SpectrometerFrequency'][0] == 127.765536
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == data_path_2.name

    # Similar data from another test set
    subprocess.check_call(['spec2nii', 'philips_dcm',
                           '-f', 'svs2',
                           '-o', tmp_path,
                           '-j',
                           str(data_path_press)])

    img_metab = read_nifti_mrs(tmp_path / 'svs2.nii.gz')
    img_ref = read_nifti_mrs(tmp_path / 'svs2_ref.nii.gz')

    assert img_metab.shape == (1, 1, 1, 1024)
    assert np.iscomplexobj(img_metab.dataobj)
    assert np.isclose(1 / img_metab.header['pixdim'][4], 2000.0)

    hdr_ext_codes = img_metab.header.extensions.get_codes()
    hdr_ext = json.loads(img_metab.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['SpectrometerFrequency'][0] == 127.770768
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == data_path_press.name
    assert hdr_ext['WaterSuppressed'] is True

    assert img_ref.shape == (1, 1, 1, 1024)
    hdr_ext_codes = img_ref.header.extensions.get_codes()
    hdr_ext = json.loads(img_ref.header.extensions[hdr_ext_codes.index(44)].get_content())
    assert hdr_ext['WaterSuppressed'] is False


def test_svs_mega(tmp_path):
    subprocess.run([
        'spec2nii', 'philips_dcm',
        '-f', 'svs_mega',
        '-o', tmp_path,
        '-j',
        data_path_mpress])

    img_metab = read_nifti_mrs(tmp_path / 'svs_mega.nii.gz')
    img_ref = read_nifti_mrs(tmp_path / 'svs_mega_ref.nii.gz')

    assert img_metab.shape == (1, 1, 1, 1024, 144, 2)
    assert np.iscomplexobj(img_metab.dataobj)
    assert np.isclose(1 / img_metab.header['pixdim'][4], 2000.0)

    hdr_ext_codes = img_metab.header.extensions.get_codes()
    hdr_ext = json.loads(img_metab.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['SpectrometerFrequency'][0] == 127.770768
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == data_path_mpress.name
    assert hdr_ext['WaterSuppressed'] is True
    assert hdr_ext['dim_5'] == 'DIM_DYN'
    assert hdr_ext['dim_6'] == 'DIM_EDIT'
    assert hdr_ext['dim_6_header'] == {'EditCondition': ['ON', 'OFF']}

    assert img_ref.shape == (1, 1, 1, 1024, 9)
    hdr_ext_codes = img_ref.header.extensions.get_codes()
    hdr_ext = json.loads(img_ref.header.extensions[hdr_ext_codes.index(44)].get_content())
    assert hdr_ext['WaterSuppressed'] is False
    assert hdr_ext['dim_5'] == 'DIM_DYN'



'''Tests for UIH dicom format conversion.

Copyright William Clarke, University of Oxford 2021
Subject to the BSD 3-Clause License.
'''

import subprocess
from pathlib import Path
import json

import numpy as np

from .io_for_tests import read_nifti_mrs

# Data paths
data_base = Path(__file__).parent / 'spec2nii_test_data' / 'UIH'

data_paths = {'svs': data_base / 'mrs_data/dicom/svs_press_te144_SVS_801/00000001.dcm',
              'csi_2d': data_base / 'mrs_data/dicom/csi_hise_te144_CSI_1201/00000000.dcm',
              'csi_3d': data_base / 'mrs_3d/dicom/csi_hise_3d_te144_CSI_1301/00000000.dcm'}


def test_svs(tmp_path):

    subprocess.check_call(['spec2nii', 'uih',
                           '-f', 'svs',
                           '-o', tmp_path,
                           '-j',
                           str(data_paths['svs'])])

    img_t = read_nifti_mrs(tmp_path / 'svs.nii.gz')

    assert img_t.shape == (1, 1, 1, 2048)
    assert np.iscomplexobj(img_t.dataobj)
    assert 1 / img_t.header['pixdim'][4] == 2000.0

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['SpectrometerFrequency'][0] == 128.207081
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == data_paths['svs'].name


def test_2D(tmp_path):

    subprocess.check_call(['spec2nii', 'uih',
                           '-f', 'csi_2d',
                           '-o', tmp_path,
                           '-j',
                           str(data_paths['csi_2d'])])

    img_t = read_nifti_mrs(tmp_path / 'csi_2d.nii.gz')

    assert img_t.shape == (16, 16, 1, 2048)
    assert np.iscomplexobj(img_t.dataobj)
    assert 1 / img_t.header['pixdim'][4] == 2000.0

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['SpectrometerFrequency'][0] == 128.207084
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == data_paths['csi_2d'].name


def test_3D(tmp_path):

    subprocess.check_call(['spec2nii', 'uih',
                           '-f', 'csi_3d',
                           '-o', tmp_path,
                           '-j',
                           str(data_paths['csi_3d'])])

    img_t = read_nifti_mrs(tmp_path / 'csi_3d.nii.gz')

    assert img_t.shape == (8, 8, 8, 1024)
    assert np.iscomplexobj(img_t.dataobj)
    assert 1 / img_t.header['pixdim'][4] == 2000.0

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['SpectrometerFrequency'][0] == 128.207089
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == data_paths['csi_3d'].name



'''Tests for spec2nii dump, extract and insert functions.

Copyright William Clarke, University of Oxford 2021
Subject to the BSD 3-Clause License.
'''

import subprocess
from pathlib import Path
import json

import numpy as np
from fsl.data.image import Image

from .io_for_tests import read_nifti_mrs

# Data paths
siemens_path = Path(__file__).parent / 'spec2nii_test_data' / 'Siemens'
data_path = siemens_path / 'VBData' / 'Twix' / 'meas_MID151_svs_se_C_T15_S10_10_FID108741.dat'


def test_dump(tmp_path):
    # Convert twix
    subprocess.check_call(['spec2nii', 'twix',
                           '-e', 'image',
                           '-f', 'original',
                           '-o', tmp_path,
                           '-j', str(data_path)])

    subprocess.check_call(['spec2nii', 'dump', str(tmp_path / 'original.nii.gz')])


def test_extract(tmp_path):
    # Convert twix
    subprocess.check_call(['spec2nii', 'twix',
                           '-e', 'image',
                           '-f', 'original',
                           '-o', tmp_path,
                           '-j', str(data_path)])

    subprocess.check_call(['spec2nii', 'extract',
                           '-o', tmp_path,
                           str(tmp_path / 'original.nii.gz')])

    assert (tmp_path / 'original.json').exists()

    with open(tmp_path / 'original.json') as jf:
        extracted_hdr = json.load(jf)

    img = read_nifti_mrs(tmp_path / 'original.nii.gz')
    hdr_ext_codes = img.header.extensions.get_codes()
    hdr_ext = json.loads(img.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert extracted_hdr == hdr_ext


def test_insert(tmp_path):
    # Convert twix
    subprocess.check_call(['spec2nii', 'twix',
                           '-e', 'image',
                           '-f', 'original',
                           '-o', tmp_path,
                           '-j', str(data_path)])

    subprocess.check_call(['spec2nii', 'extract',
                           '-o', tmp_path,
                           str(tmp_path / 'original.nii.gz')])

    assert (tmp_path / 'original.json').exists()

    with open(tmp_path / 'original.json') as jf:
        extracted_hdr = json.load(jf)

    extracted_hdr['SpectrometerFrequency'] = [123.456789, ]

    with open(tmp_path / 'new.json', 'w') as jf:
        json.dump(extracted_hdr, jf)

    subprocess.check_call(['spec2nii', 'insert',
                           '-o', tmp_path,
                           '-f', 'new',
                           '--dwelltime', '0.0005',
                           str(tmp_path / 'original.nii.gz'),
                           str(tmp_path / 'new.json')])

    img = read_nifti_mrs(tmp_path / 'new.nii.gz')
    hdr_ext_codes = img.header.extensions.get_codes()
    hdr_ext = json.loads(img.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert extracted_hdr == hdr_ext
    assert np.isclose(img.header['pixdim'][4], 0.0005)


def test_insert_not_nmrs(tmp_path):
    # Create a small dummy file that doesn't conform to the standard

    affine = np.random.random((4, 4))
    affine[3, 0] = 0.0
    affine[3, 1] = 0.0
    affine[3, 2] = 0.0
    affine[3, 3] = 1.0

    data = np.ones((3, 3, 3, 256), dtype=float) + 1j * np.ones((3, 3, 3, 256), dtype=float)

    non_compliant = Image(data, header=None, xform=affine)
    non_compliant.header['pixdim'][4] = 1 / 1000
    non_compliant.save(tmp_path / 'noncompliant.nii.gz')

    newhd = {
        'SpectrometerFrequency': 123.456789,
        'ResonantNucleus': '1H'}

    with open(tmp_path / 'new.json', 'w') as jf:
        json.dump(newhd, jf)

    subprocess.check_call(['spec2nii', 'insert',
                           '-o', tmp_path,
                           '-f', 'compliant',
                           str(tmp_path / 'noncompliant.nii.gz'),
                           str(tmp_path / 'new.json')])

    img = read_nifti_mrs(tmp_path / 'compliant.nii.gz')
    hdr_ext_codes = img.header.extensions.get_codes()
    hdr_ext = json.loads(img.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert img.header.get_intent()[2].split('_')[0] == 'mrs'
    assert np.allclose(img.affine, affine)
    assert np.isclose(img.header['pixdim'][4], 1 / 1000)

    subprocess.check_call(['spec2nii', 'insert',
                           '-o', tmp_path,
                           '-f', 'compliant_dt',
                           '--dwelltime', '0.0005',
                           str(tmp_path / 'noncompliant.nii.gz'),
                           str(tmp_path / 'new.json')])

    img = read_nifti_mrs(tmp_path / 'compliant_dt.nii.gz')
    assert np.isclose(img.header['pixdim'][4], 0.0005)



'''Tests for NIfTI-MRS anonymisation.

Copyright William Clarke, University of Oxford 2021
Subject to the BSD 3-Clause License.
'''

import numpy as np
import subprocess
from .io_for_tests import read_nifti_mrs
from pathlib import Path
import json

# Data paths
siemens_path = Path(__file__).parent / 'spec2nii_test_data' / 'Siemens'
data_path = siemens_path / 'VBData' / 'Twix' / 'meas_MID151_svs_se_C_T15_S10_10_FID108741.dat'


def test_anon(tmp_path):
    # Convert twix
    subprocess.check_call(['spec2nii', 'twix',
                           '-e', 'image',
                           '-f', 'original',
                           '-o', tmp_path,
                           '-j', str(data_path)])

    subprocess.check_call(['spec2nii', 'anon',
                           '-f', 'anon',
                           '-o', tmp_path,
                           '-r', 'PatientSex',
                           '-r', 'RepetitionTime',
                           '-v', str(tmp_path / 'original.nii.gz')])

    img_o = read_nifti_mrs(tmp_path / 'original.nii.gz')
    img_a = read_nifti_mrs(tmp_path / 'anon.nii.gz')

    assert img_o.shape == img_a.shape
    assert np.allclose(img_o.get_fdata(dtype=complex), img_a.get_fdata(dtype=complex))
    assert np.allclose(img_o.affine, img_a.affine)
    assert np.isclose(img_o.header['pixdim'][4], img_a.header['pixdim'][4])

    hdr_ext_codes = img_o.header.extensions.get_codes()
    hdr_ext_o = json.loads(img_o.header.extensions[hdr_ext_codes.index(44)].get_content())
    hdr_ext_codes = img_a.header.extensions.get_codes()
    hdr_ext_a = json.loads(img_a.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert 'SpectrometerFrequency' in hdr_ext_a
    assert 'ResonantNucleus' in hdr_ext_a
    assert 'dim_5' in hdr_ext_a
    assert 'OriginalFile' in hdr_ext_o
    assert 'OriginalFile' not in hdr_ext_a
    assert 'PatientName' in hdr_ext_o
    assert 'PatientName' not in hdr_ext_a
    assert 'PatientSex' not in hdr_ext_a
    assert 'RepetitionTime' not in hdr_ext_a


def test_inline_anon(tmp_path):
    # Convert twix
    subprocess.check_call(['spec2nii', 'twix',
                           '-e', 'image',
                           '-f', 'original',
                           '-o', tmp_path,
                           '-j', str(data_path),
                           '--anon',
                           '--verbose'])

    img_o = read_nifti_mrs(tmp_path / 'original.nii.gz')

    hdr_ext_codes = img_o.header.extensions.get_codes()
    hdr_ext_o = json.loads(img_o.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert 'SpectrometerFrequency' in hdr_ext_o
    assert 'ResonantNucleus' in hdr_ext_o
    assert 'dim_5' in hdr_ext_o
    assert 'OriginalFile' not in hdr_ext_o
    assert 'PatientName' not in hdr_ext_o
    assert 'PatientDoB' not in hdr_ext_o
